// $Id: op_new_t.cc 2742 2022-09-17 21:59:20Z teshields $

// This file is part of the Libnew package that provides replacement
// implementations for core standard C library memory management
// functions, and for the standard C++ 'new()' and 'delete()' operators.
//
// Copyright (C) 2020-2022  Thomas E. Shields
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#if defined (TRACE)
#include <iostream>
#endif

#include <new>

#if defined (USE_BDWGC)
#include <gc/gc.h>
#elif defined (USE_DMALLOC)
#include <dmalloc.h>
#endif

void* operator new (std::size_t sz)
#if __cplusplus >= 201703L // >= C++17
                                    noexcept (false)
#elif defined (_GLIBCXX_THROW)
                                    _GLIBCXX_THROW (std::bad_alloc)
#else
                                    throw (std::bad_alloc)
#endif
{
  void* p;

#if defined (TRACE)
  std::cerr << "operator new (size_t) throw (bad_alloc)"
               ": allocating " << sz << " bytes";
#endif

  if (sz == 0) sz = 1; // treat 0-byte request as a 1-byte request

  // attempt to allocate size bytes;
  while ((p =
#if defined (USE_BDWGC)
              GC_MALLOC (sz)
#elif defined (USE_DMALLOC)
              dmalloc_malloc (__FILE__, __LINE__, sz, DMALLOC_FUNC_MALLOC, 0, 0)
#else
              malloc (sz)
#endif
         ) ==
#if __cplusplus >= 201103L // >= C++11
              nullptr
#else
              NULL
#endif
        )
    {
      // allocation was unsuccessful;
      // get the current new handler function
#if __cplusplus >= 201103L && HAVE_GET_NEW_HANDLER // >= C++11 w/ get_new_handler ()
      std::new_handler handler = std::get_new_handler ();
#else
      std::new_handler handler = std::set_new_handler (
#if __cplusplus >= 201103L // >= C++11
                                                       nullptr
#else
                                                       NULL
#endif
                                                      );
      std::set_new_handler (handler);
#endif

      if (handler ==
#if __cplusplus >= 201103L // >= C++11
                     nullptr
#else
                     NULL
#endif
         )                        // if no handler is registered
         throw std::bad_alloc (); // throw an exception
      handler ();                 // hander is registered, call it
    }

  // the allocation was successful
#if defined (TRACE)
  std::cerr << " @ " << p << std::endl;
#endif
  return p;                     // return a pointer to the memory
}
