/* $Id: code.c 2579 2022-06-23 17:54:48Z teshields $ */
  
/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1993  Kurt M. Bischoff
   Copyright (C) 2016-2022  Thomas E. Shields
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "gppl_glob.h"
#include "code.h"
#include "parser.h"
#include "util.h"

#define ceiling(num,inc) (((inc) * ((num)/(inc))) + (((num)%(inc)!=0)?(inc):0))

#define ALIGNSIZE   4
#define STACKSIZE1 500000L
#define STACKSIZE  (ceiling(STACKSIZE1,ALIGNSIZE)) 

#define nFREGS 3 
#define RETREGADDR1   -2    /* relative to the callee's fp */
#define RETREGADDR2   -4    /* relative to sp upon return  */ 

/* RA, VA, PA are for calculations of frame-pointer-relative addresses */
#define RA(regNum)       (-(nFREGS + (regNum)))
#define VA(variableAddr) (-(nFREGS + dumEnv->nRegs + (variableAddr))) 
#define PA(paramAddr)    ((paramAddr) + 1 + 1)


enum pred doCodeGen = True;

FILE *tempOut; 



long stgSize(enum symType type, long howMany) 
  {return(howMany);
  } 



static labelType labelSeq = 0; 
static labelType returnLabelSeq = 0; 



labelType newLabelNum(void) 
  {return (labelSeq++); 
  }



labelType newReturnLabelNum(void) 
  {return (returnLabelSeq++); 
  }

 

static const char *cHeaderCode[] =
{ 
"\n",
"#include <stdlib.h>\n",
"#include <stdio.h>\n",
"\n",
"union stackItem {\n",
"   int i; \n",
"   float f;\n",
"   int b; \n",
"   int *ip;\n",
"   float *fp;\n",
"   int *bp; \n",
"   union stackItem *up;\n",
"  };\n",
"\n",
"\n",
"int main()\n",
"{\n",
" register union stackItem *sp,*fp;\n",
"\n",
" if ((sp = (union stackItem *)malloc(STACKSIZE * sizeof(union stackItem)))\n",
"     ==\n",
"     ((union stackItem *)NULL)\n",
"    )\n",
"    fprintf(stderr,\"malloc error in stack allocation\\n\");\n",
" sp = sp + STACKSIZE;\n",
"\n",
}; 



void cgHeader(void) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"#define STACKSIZE %ld\n",STACKSIZE); 
   printStringArray(cHeaderCode,(int)sizeof(cHeaderCode),tempOut); 
  }



void cgTrailer(void) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\n\texit(0);\n");  
   fprintf(tempOut,"rts:\n");
   fprintf(tempOut,"  switch ((*(sp++)).i)  \n    {\n");
   for (labelType i=0; i<returnLabelSeq; i++) 
       fprintf(tempOut,"     case %4d: goto r%d;\n",(int)i,(int)i); 
   fprintf(tempOut,"    } /* rts switch */ \n"); 
   fprintf(tempOut,"IOBtrap:\n"); 
   fprintf(tempOut,"\tfprintf(stderr,\"array index out of bounds.\\n\");\n"); 
   fprintf(tempOut,"\texit(-1);\n"); 
   fprintf(tempOut,"} /* main */ \n"); 
  }



void cgFuncCall(struct env *env,
		struct sym *funcSym,
		regNumType regNum,
		enum pred isEmptyActList,
		lineNumType line) 
  {labelType retLabel; 
   levelType staticDist;
   struct sym *dum;

   if ((doCodeGen == False) || (funcSym == lambdaSym)) return; 
   if (isEmptyActList == True)
      {dum = funcSym->paramLoc;
       while (dum != lambdaSym)
         {fprintf(stderr,"missing actual parameter in line %d's call to %s.\n",
                         (int)line,funcSym->id
                 ); 
          handleSemError();
          dum = dum->next;
         }
      }
   staticDist = env->level - funcSym->level;
   retLabel = newReturnLabelNum(); 
   fprintf(tempOut,"\t(*--sp).i = %d;\n",(int)retLabel);  
   fprintf(tempOut,"\t(*--sp).up = fp;\n");  
   for (levelType i=0; i<staticDist; i++) 
       fprintf(tempOut,"\t*sp = *(*sp).up;\n");  
   fprintf(tempOut,"\tgoto f%d;\n",(int)(funcSym->label));  
   fprintf(tempOut,"r%d:\n",(int)retLabel); 
   if (funcSym->type != VOIDT)
      fprintf(tempOut,"\tfp[%d] = sp[%d];\n",RA(regNum),RETREGADDR2); 
   fprintf(tempOut,"\tsp += %ld;\n",funcSym->paramsSize);  
  }



void cgBlockEnter(struct env *env) 
  {if (doCodeGen == False) return; 
   if (env != lambdaEnv) 
      {if (env->funcSym == lambdaSym) 
          fprintf(tempOut,"\t(*--sp).up = fp;\n"); 
       fprintf(tempOut,"\t(*--sp).up = fp;\n"); 
       fprintf(tempOut,"\tfp = sp + 1;\n"); 
       fprintf(tempOut,"\tsp -= %ld;\n",
               1 + env->nRegs + env->varsSize 
              ); 
      }
  }



void cgBlockExit(struct env *env) 
  {if (doCodeGen == False) return; 
   if (env == lambdaEnv) return;
   fprintf(tempOut,"BE%d:\tsp = fp;\n",(int)(env->exitLabel)); 
   fprintf(tempOut,"\tfp = (*(sp++ - 1)).up;\n"); 
  }



void cgFuncAddress(struct sym *sym) 
  {if (doCodeGen == False) return; 
   if (sym != lambdaSym) 
      {fprintf(tempOut,"\tgoto j%d;\n",(int)(sym->label)); 
       fprintf(tempOut,"f%d:\t\t\t\t\t\t/* %s */\n",(int)(sym->label),sym->id); 
      }
  }



void cgFuncTrailer(struct sym *sym) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\tgoto rts;\n"); 
   fprintf(tempOut,"j%d:\n",(int)(sym->label)); 
  }



void cgBinArithExpr(int destReg, int op, enum symType type) 
  {const char *typeExt = "";  

   if (doCodeGen == False) return; 
   switch(type) 
     {case INTT:    typeExt = "i";  break;
      case FLOATT:  typeExt = "f";  break;
      default:                      break;
     }
   fprintf(tempOut,"\tfp[%d].%s %c= fp[%d].%s;\n",
           RA(destReg),
           typeExt,
           op,
           RA(destReg + 1),
           typeExt 
          );
  }



void cgRelExpr(int destReg, int op, enum symType type) 
  {const char *typeExt = "",*opStr = "";  
   labelType label;

   if (doCodeGen == False) return; 
   switch(type) 
     {case INTT:    typeExt = "i";  break;
      case FLOATT:  typeExt = "f";  break;
      default:                      break;
     }
   switch (op) 
     {case EQ:  opStr = "=="; break;
      case LEQ: opStr = "<="; break;
      case GEQ: opStr = ">="; break;
      case NEQ: opStr = "!="; break;
      case LT:  opStr = "<"; break;
      case GT:  opStr = ">"; break;
     } 
   label = newLabelNum(); 
   fprintf(tempOut,"\tif (fp[%d].%s %s fp[%d].%s) goto b%d;\n",
           RA(destReg),typeExt,opStr,
           RA(destReg + 1),typeExt,(int)label
          );
   fprintf(tempOut,"\tfp[%d].i = 0;\n",RA(destReg)); 
   fprintf(tempOut,"\tgoto B%d;\n",(int)label); 
   fprintf(tempOut,"b%d:\tfp[%d].i = 1;\n",(int)label,RA(destReg)); 
   fprintf(tempOut,"B%d:\n",(int)label); 
  }



void cgBinLogExpr(int destReg, int op) 
  {labelType label;
   char unrOp = '\0',rslt1 = '\0',rslt2 = '\0'; 

   if (doCodeGen == False) return; 
   label = newLabelNum(); 
   switch(op)
     {case AND: unrOp = '!'; rslt1 = '1'; rslt2 = '0'; break;  
      case OR : unrOp = ' '; rslt1 = '0'; rslt2 = '1'; break;  
     }
   fprintf(tempOut,"\tif (%c(fp[%d].i)) goto b%d;\n",
           unrOp,RA(destReg),(int)label
          );
   fprintf(tempOut,"\tif (%c(fp[%d].i)) goto b%d;\n",
           unrOp,RA(destReg + 1),(int)label
          );
   fprintf(tempOut,"\tfp[%d].i = %c;\n",RA(destReg),rslt1); 
   fprintf(tempOut,"\tgoto B%d;\n",(int)label); 
   fprintf(tempOut,"b%d:\tfp[%d].i = %c;\n",(int)label,RA(destReg),rslt2); 
   fprintf(tempOut,"B%d:\n",(int)label); 
  }



void cgUnrArithExpr(int destReg, int op, enum symType type) 
  {const char *typeExt = "";  

   if (doCodeGen == False) return; 
   if (op == '+') return; 
   switch(type) 
     {case INTT:    typeExt = "i";  break;
      case FLOATT:  typeExt = "f";  break;
      default:                      break;
     }
   fprintf(tempOut,"\tfp[%d].%s = -fp[%d].%s;\n",
           RA(destReg),typeExt,RA(destReg),typeExt 
          );
  }



void cgLogNegExpr(int regNum) 
  {labelType label;
   
   if (doCodeGen == False) return; 
   label = newLabelNum(); 
   fprintf(tempOut,"\tif ((fp[%d].i) == 0) goto b%d;\n",
           RA(regNum),(int)label
          );
   fprintf(tempOut,"\tfp[%d].i = 0;\n",RA(regNum)); 
   fprintf(tempOut,"\tgoto B%d;\n",(int)label); 
   fprintf(tempOut,"b%d:\tfp[%d].i = 1;\n",(int)label,RA(regNum)); 
   fprintf(tempOut,"B%d:\n",(int)label); 
  }



void cgEmptyIndexSeq(struct env *env,
		     struct sym *sym,
		     lineNumType line,
		     int regNum,
		     struct dim *dimList) 
  {long addr; 
   levelType staticDist;
   struct env *dumEnv;
   
   if (doCodeGen == False) return; 
   if (sym == lambdaSym) return; 
   staticDist = env->level - sym->level;
   while (dimList != lambdaDim) 
      {fprintf(stderr,
               "missing index for line %d\'s reference to array %s.\n",
               (int)line,sym->id
              ); 
       dimList = dimList->next;
       handleSemError();   
      }
   fprintf(tempOut,"\tfp[%d].up = fp;\n",RA(regNum)); 
   dumEnv = env; 
   for (levelType i=0; i<staticDist; i++) 
       {dumEnv = dumEnv->outerEnv;
        fprintf(tempOut,"\tfp[%d] = *(fp[%d].up);\n",RA(regNum),RA(regNum)); 
       }
   addr = (sym->Class == VARIABLEC)?(VA(sym->addr)):(PA(sym->addr));
   fprintf(tempOut,"\tfp[%d].up += %ld;\n",RA(regNum),addr); 
   if (sym->Class == REFPARAMC)
      fprintf(tempOut,"\tfp[%d] = *(fp[%d].up);\n",
                      RA(regNum),RA(regNum) 
             );
  }



void cgIndexSeq(struct sym *sym, int regNum, struct dim *dimList) 
  {if (doCodeGen == False) return; 
   if (sym == lambdaSym) return; 
   fprintf(tempOut,"\tif (fp[%d].i < 0) goto IOBtrap;\n",
           RA(regNum + 1)
          ); 
   fprintf(tempOut,"\tif (fp[%d].i > %ld) goto IOBtrap;\n",
           RA(regNum + 1),
           dimList->size - 1
          ); 
   fprintf(tempOut,"\tfp[%d].i *= %ld;\n",RA(regNum + 1),dimList->scale); 
   fprintf(tempOut,"\tfp[%d].up -= fp[%d].i;\n",
           RA(regNum),RA(regNum + 1)
          ); 
  }



void cgConstSimpleExpr(int regNum, enum symType type, union num value) 
  {if (doCodeGen == False) return; 
   switch(type)
     {case INTT:    
        fprintf(tempOut,"\tfp[%d].i = %d;\n",RA(regNum),value.i);  
        break;
      case FLOATT:  
        fprintf(tempOut,"\tfp[%d].f = %f;\n",RA(regNum),(double)value.f);  
        break;
      case BOOLT:   
        fprintf(tempOut,"\tfp[%d].b = %d;\n",RA(regNum),value.b);  
        break;
      default:
        break;
     }
  }



void cgEOF(int regNum) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\tfp[%d].i = EOF;\n",RA(regNum));  
  }



void cgAssign(int regNum, enum symType type) 
  {const char *typeExt = "";
   
   if (doCodeGen == False) return; 
   switch(type) 
     {case INTT:    typeExt = "i";  break;
      case FLOATT:  typeExt = "f";  break;
      case BOOLT:   typeExt = "b";  break;
      default:                      break;
     }
   fprintf(tempOut,"\t*(fp[%d].%sp) = fp[%d].%s;\n",
           RA(regNum),typeExt,RA(regNum + 1),typeExt
          ); 
   fprintf(tempOut,"\tfp[%d].%s = fp[%d].%s;\n",
           RA(regNum),typeExt,RA(regNum + 1),typeExt
          ); 
  }



void cgPush(int regNum) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\t*--sp = fp[%d];\n",RA(regNum));  
  }



void cgCondDeref(struct sym *formal, int regNum, enum symType type) 
  {const char *typeExt = "";

   if (doCodeGen == False) return; 
   if ((formal != lambdaSym) && (formal->Class == REFPARAMC)) return;
   switch(type) 
     {case INTT:    typeExt = "i";  break;
      case FLOATT:  typeExt = "f";  break;
      case BOOLT:   typeExt = "b";  break;
      default:                      break;
     }
   fprintf(tempOut,"\tfp[%d].%s = *(fp[%d].%sp);\n",
           RA(regNum),typeExt,
           RA(regNum),typeExt
          ); 
  }



void cgCast(int regNum, enum symType newType, enum symType oldType) 
  {const char *ext1 = "",*ext2 = "",*cCast = "";
  
   if (doCodeGen == False) return; 
   switch(newType) 
     {case INTT:    ext1 = "i"; cCast = "(int)";   break; 
      case FLOATT:  ext1 = "f"; cCast = "(float)"; break; 
      case BOOLT:   ext1 = "b"; cCast = "(int)";   break; 
      default:                                     break;
     }
   switch(oldType) 
     {case INTT:    ext2 = "i";  break;
      case FLOATT:  ext2 = "f";  break;
      case BOOLT:   ext2 = "b";  break;
      default:                   break;
     }
   fprintf(tempOut,"\tfp[%d].%s = %sfp[%d].%s;\n",
           RA(regNum),ext1,cCast,RA(regNum),ext2
          );
  }



void cgOutputNoArg(char *str, regNumType regNum) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\tfp[%d].i = printf(%s);\n",RA(regNum),str);
    /*  fprintf(tempOut,"\tfflush(stdout);\n"); */
  }



void cgOutputOneArg(char *str, int regNum, enum symType type) 
  {if (doCodeGen == False) return; 
   switch(type) 
     {case INTT: 
        fprintf(tempOut,"\tfp[%d].i = printf(%s,fp[%d].i);\n",
                RA(regNum),str,RA(regNum));
        break;
      case FLOATT:    
        fprintf(tempOut,
                "\tfp[%d].i = printf(%s,fp[%d].f);\n",
                RA(regNum),str,RA(regNum)
               );
        break;
      case BOOLT:   
        fprintf(tempOut,
          "\tfp[%d].i = printf(%s,(*(fp[%d].b)?\"true\":\"false\"));\n",
                RA(regNum),str,RA(regNum)
               );
        break;
      default:
        break;
     }
   /*  fprintf(tempOut,"\tfflush(stdout);\n"); */
  }



void cgInput(char *str, int regNum, enum symType type, lineNumType line) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\tfp[%d].i = ",RA(regNum));
   switch(type) 
     {case INTT: 
        fprintf(tempOut,"scanf(%s,fp[%d].ip);\n",
                str,RA(regNum));
        break;
      case FLOATT:    
        fprintf(tempOut,"scanf(%s,fp[%d].fp);\n",
                str,RA(regNum));
        break;
      case BOOLT:   
        fprintf(stderr,"attempted input of boolean near line %d.\n",(int)line); 
        handleSemError(); 
        break;
      default:
        break;
     }
  }



static void condCheck(enum symType type, lineNumType line)
  {if (type != BOOLT) 
      {fprintf(stderr,
   "non-boolean expression as condition in control statement near line %d.\n",
               (int)line
              );
       handleSemError();  
      } 
  } 



void cgIfThen(enum symType type, lineNumType line, labelType label) 
  {condCheck(type,line); 
   if (doCodeGen == False) return; 
   fprintf(tempOut,"e%d:\n",(int)label); 
  }



void cgIfThenElse(enum symType type, lineNumType line, labelType label) 
  {condCheck(type,line); 
   if (doCodeGen == False) return; 
   fprintf(tempOut,"i%d:\n",(int)label); 
  }



void cgThen(int regNum, labelType label) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\tif (!(fp[%d].i)) goto e%d;\n",RA(regNum),(int)label); 
  }



void cgElse(labelType label) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"kl%d:\tgoto i%d;\n",(int)label,(int)label); 
   fprintf(tempOut,"e%d:\n",(int)label); 
   /* labels prefixed by kl are kludges to silence the C compiler's
    *   "statement not reached" warnings when the then clause ends
    *   with a return statement.
    */
  }



void cgWdecision(int regNum, labelType label) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\tif (!(fp[%d].i)) goto W%d;\n",RA(regNum),(int)label); 
  }



void cgWhileStatement(enum symType type, lineNumType line, labelType label) 
  {condCheck(type,line); 
   if (doCodeGen == False) return; 
   fprintf(tempOut,"\tgoto w%d;\n",(int)label); 
   fprintf(tempOut,"W%d:\n",(int)label); 
  }



void cgWhile(labelType label) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"w%d:\n",(int)label); 
  }



void cgReturnStmt(struct env *env,
		   regNumType regNum,
		   enum symType type,
		   lineNumType line) 
  {if (doCodeGen == False) goto err; 
   if (type != VOIDT) 
      fprintf(tempOut,"\tfp[%d] = fp[%d];\n",RETREGADDR1,RA(regNum)); 
   while (env->funcSym == lambdaSym) 
     {fprintf(tempOut,"\tsp = fp;\n"); 
      fprintf(tempOut,"\tfp = (*fp).up;\n"); 
      if (type != VOIDT) 
         fprintf(tempOut,"\tfp[%d] = sp[%d];\n",RETREGADDR1,RETREGADDR1); 
      if ((env = env->outerEnv) == lambdaEnv)
         {fprintf(stderr,"line %d: return from no function.\n",(int)line); 
          goto err; 
         }
     }
   if (env->funcSym->type != VOIDT)  
      {if (type == VOIDT)
          {fprintf(stderr,
           "line %d: missing object of return in non-void function %s.\n",
                   (int)line,env->funcSym->id 
                  ); 
           goto err; 
          }
       if (env->funcSym->type != type) 
          {fprintf(stderr,
           "line %d: mismatched type of object returned from function %s.\n",
                  (int)line,env->funcSym->id 
                 ); 
           goto err; 
          }
      }
      else
      {if (type != VOIDT)
          {fprintf(stderr,"line %d: object returned by void function %s.\n",
                   (int)line,env->funcSym->id 
                  ); 
           goto err; 
          }
      }
   fprintf(tempOut,"\tgoto BE%d;\n",(int)(env->exitLabel)); 
   return;
err:
   handleSemError(); 
   return;
  }



void cgExitStmt(regNumType regNum, enum symType type, lineNumType line) 
  {if (doCodeGen == False) return; 
   if (type != INTT) 
      {fprintf(stderr,"line %d: non-integer object of exit statement.\n",
               (int)line
              );
       handleSemError();
       return;
      }
   fprintf(tempOut,"\texit(fp[%d].i);\n",RA(regNum)); 
  }



