%{
/* $Id: gppl.L 2539 2022-05-23 00:29:05Z teshields $ */ 

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.
 
   Copyright (C) 1992-1993  Kurt M. Bischoff
   Copyright (C) 2016-2022  Thomas E. Shields
 
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "gppl_glob.h" 
#include "scan.h"
#include "parser.h" 

%}

IDENT ([a-zA-Z_][a-zA-Z0-9_]*) 

%% 
int            return(INT);        @{ @INT.line@     = (lineNumType)yylineno; @}  
float          return(TKFLOAT);    @{ @TKFLOAT.line@ = (lineNumType)yylineno; @} 
bool           return(BOOL);       @{ @BOOL.line@    = (lineNumType)yylineno; @} 
void           return(VOID);       @{ @VOID.line@    = (lineNumType)yylineno; @} 
var            return(VAR); 
if             return(IF);         @{ @IF.line@      = (lineNumType)yylineno; @} 
then           return(THEN); 
else           return(ELSE); 
while          return(WHILE);      @{ @WHILE.line@   = (lineNumType)yylineno; @} 
return         return(RETURN);     @{ @RETURN.line@  = (lineNumType)yylineno; @} 
exit           return(EXIT);       @{ @EXIT.line@    = (lineNumType)yylineno; @} 
input          return(INPUT); 
output         return(OUTPUT); 
\"             {char *strDum;
                strDum = copyLexemeOfSTRING(); 
                return (STRING);  
                                   @{ @STRING.str@   = strDum;   @}
               }
\/\*           eatComment(); 
[ \n\t\f]*     ;  
[0-9]+\.[0-9]+ return(FCONST);     @{ (void)sscanf((char *)yytext,"%f",
                                                   &(@FCONST.val@.f)
                                                  ); 
                                   @}
[0-9]+         return(ICONST);     @{ (void)sscanf((char *)yytext,"%d",
                                                   &(@ICONST.val@.i)
                                                  ); 
                                   @}
EOF            return(EOFCONST);  
TRUE           return(BCONST);     @{ @BCONST.val@.b = 1;        @} 
FALSE          return(BCONST);     @{ @BCONST.val@.b = 0;        @} 
\(             return('(');        @{ @'('.line@     = (lineNumType)yylineno; @}
\[             return('[');        @{ @'['.line@     = (lineNumType)yylineno; @}
[,]            return(',');        @{ @','.line@     = (lineNumType)yylineno; @}
[{]            return('{');
[}]            return('}');
[)]            return(')');
[;]            return(';');
[\]]           return(']');
&&             return(AND);        @{ @AND.line@     = (lineNumType)yylineno; @} 
\|\|           return(OR);         @{ @OR.line@      = (lineNumType)yylineno; @} 
!              return(NOT);        @{ @NOT.line@     = (lineNumType)yylineno; @} 
[*]            return('*');        @{ @'*'.line@     = (lineNumType)yylineno; @} 
[+]            return('+');        @{ @'+'.line@     = (lineNumType)yylineno; @} 
[/]            return('/');        @{ @'/'.line@     = (lineNumType)yylineno; @} 
[\-]           return('-');        @{ @'-'.line@     = (lineNumType)yylineno; @} 
=              return(GETS);       @{ @GETS.line@    = (lineNumType)yylineno; @} 
==             return(EQ);         @{ @EQ.line@      = (lineNumType)yylineno; @} 
!=             return(NEQ);        @{ @NEQ.line@     = (lineNumType)yylineno; @} 
"<"            return(LT);         @{ @LT.line@      = (lineNumType)yylineno; @} 
">"            return(GT);         @{ @GT.line@      = (lineNumType)yylineno; @} 
"<="           return(LEQ);        @{ @LEQ.line@     = (lineNumType)yylineno; @} 
">="           return(GEQ);        @{ @GEQ.line@     = (lineNumType)yylineno; @} 
{IDENT}        {
                return(ID);   
                @{ @ID.str@ = stringLookup((char *)yytext,yyleng,1); 
                   @ID.line@ = (lineNumType)yylineno; 
                @} 
               } 
.              {fprintf(stderr,"line %d: illegal character\n",yylineno); 
                handleSemError();
               }
%%
#if (!defined(FLEX_SCANNER) && !defined(REFLEX_VERSION)) || !defined(__cplusplus)
int yyyinput(void){return input();} /* `lex', or `flex' with C compiler */
#else
int yyyinput(void){return yyinput();} /* `flex' with C++ compiler. or `reflex' */
#endif

int yywrap(void){return 1;}
