/* $Id: depGs.cc 2841 2022-12-13 01:08:03Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2010-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <iomanip>
#include <ostream>
#include <sstream>

#include <list>
#include <queue>
#include <vector>

#include "location.hh"

#include "codegen.hh"
#include "ctc.hh"
#include "depGs.hh"
#include "fileMgr.hh"
#include "joinTabs.hh"
#include "messages.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "scan.hh"
#include "semtypes.hh"
#include "symTrans.hh"
#include "trav.hh"
#include "treeviz.hh"


#define NUMTERMSINCR 1024
static std::vector<refCount> leafRefCount (NUMTERMSINCR,
                                           refCount ());



long eeCountInitSeq = 0;
long attrbOccrSeq = 0;
long leafRefCountInitSeq = 0;
long entOccrSeq = 0;

bool permitCtimeCircTest = true;



// an increasing sequence of numbers of grammar symbol occurrences
#define firstGSoccur 0L
long gsOccurSeq = firstGSoccur;



static int dependentSeq = 0, // count of dependents
           dependeeSeq = 0;  // count of dependees

#define NUMPERDEPSPECINCR 128

// At the time I designed this, it seemed that it would be handy to store
// both the local dependency graph and its inverse.  As I write this
// comment, I don't know if they are both used.  --KB
// The comments refer to both graphs as the LDG.

// the space for storing lists of dependents in the LDG
static std::vector<attrbPair> dependent (NUMPERDEPSPECINCR, attrbPair ());

// the space for storing lists of dependees in the LDG
static std::vector<attrbPair> dependee (NUMPERDEPSPECINCR, attrbPair ());



bool hasYaccPvarRefGlob;

// Each `struct ldgAdjListHeads' corresponds to an attribute occurrence in
// the rule being parsed.
// The following are predicates on the attribute occurrences of the rule:
//
//   defdThisProd:    is defined in this rule?  i.e., is an output occurrence?
//   hasYaccPvarRef:  definition has a reference to a Yacc pseudo variable?
//   inhFellowEnt:    has a fellow dependent that is inherited?
//   synFellowEnt:    has a fellow dependent that is synthesized?
//                    (fellow dependents are possible using the explicit
//                     mode annunciator `@e')

typedef std::list<attrbPair> attrbPairList;

struct ldgAdjListHeads
{
  bool defdThisProd;
  bool hasYaccPvarRef;
  bool inhFellowEnt;
  bool synFellowEnt;
  attrbRefSeg attrbRefText;
  attrbPairList entList; // the occurrence's dependents
  attrbPairList eeList;  // the occurrence's dependees
  int eeCount;           // the occurrence's dependee count

  ldgAdjListHeads ()
      : defdThisProd (false),
        hasYaccPvarRef (false),
        inhFellowEnt (false),
        synFellowEnt (false),
        attrbRefText (),
        entList (attrbPairList ()),
        eeList (attrbPairList ()),
        eeCount (-1)
  {}
};

static std::vector<ldgAdjListHeads> ldgALhead;

// index of pointers to the first attribute occurrence in each symbol
static std::vector<long> ldgALheadInd;



// Used in the consistency and completeness tests.
// if (change is nonzero)
//     mark the occurrence as defined or undefined (depending on newVal) ;
// return the truth value of "was it defined before calling this?".
static bool
markOrCheckAttrbOccrDefd (attrbPair attrbOccur, bool change, bool newVal)
{
  bool prev;

  prev = ldgALhead[ldgALheadInd[attrbOccur.pos] + attrbOccur.num].defdThisProd;
  if (change)
    ldgALhead[ldgALheadInd[attrbOccur.pos] + attrbOccur.num].defdThisProd =
        newVal;
  return (prev);
}



bool decorationRequested;

static std::queue<attrbPair> ldgTopSortQ;



// Inserts an occurrence into the image of the dependency expression.
// Searches the dependee or dependent list (as specified by entNOTee)
// and remembers whether the occurrence is already in that list.
// The sorting and binary search are probably overkill, since Ox
// presently is I/O-bound.
// Inserts the occurrence into the selected list if (insert == true).
// Returns true if the occurrence was already in the given list.
static void
memberDependImpl (std::vector<attrbPair>& occrList,
                  int& seq,
                  bool& member,
                  attrbPair attrbOccur,
                  bool insert)
{
  int lower = 0, upper = seq - 1;
  const int limit = (int)occrList.size ();

  while (lower <= upper)
    {
      int i = (lower + upper) / 2;
      if (attrbOccur.pos == occrList[i].pos &&
          attrbOccur.num == occrList[i].num)
        {
          member = true;
          break;
        }
      else
        {
          if (attrbOccur.pos < occrList[i].pos ||
              (attrbOccur.pos == occrList[i].pos &&
               attrbOccur.num < occrList[i].num))
            lower = i + 1;
          else
            upper = i - 1;
        }
    }
  if (insert && !member)
    {
      if (limit == seq)
        try
          {
            occrList.resize (limit + NUMPERDEPSPECINCR, attrbPair ());
          }
        catch (std::bad_alloc const&)
          {
            // reallocation of dependent/dependee list failed
            yox_alloc_fail (__FILE__, __LINE__);
          }
      for (int i = seq; i > lower; i--)
        occrList[i] = occrList[i - 1];
      occrList[lower] = attrbOccur;
      seq++;
    }
}

bool
memberDepend (attrbPair attrbOccur, bool entNOTee, bool insert)
{
  bool member = false;

  if ((attrbOccur.pos == -1) || (attrbOccur.num == -1))
    return (false);

  if (entNOTee)
    {
      if (insert)
        member = markOrCheckAttrbOccrDefd (attrbOccur, true, true);
      memberDependImpl (dependent, dependentSeq, member, attrbOccur, insert);
    }
  else
    memberDependImpl (dependee, dependeeSeq, member, attrbOccur, insert);

  return (member);
}



// get ready to start a new local dependency graph: initialize the
// local dependency graph data structures
void
initLDG (void)
{

  try
    {
      ldgALheadInd.resize (prodBufferPointer2 + 1, 0l);
    }
  catch (std::bad_alloc const&)
    {
      // resize of the local dependency graph head index failed
      yox_alloc_fail (__FILE__, __LINE__);
    }

  // clear out any existing entries
  for (long& entry : ldgALheadInd)
    entry = 0l;

  int ldgALheadIndSeq = 0;

  for (int i = 0; i < prodBufferPointer2; i++)
    {
      ldgALheadInd[i] = ldgALheadIndSeq;
      ldgALheadIndSeq += gSymTable[prodBuffer2[i].lexeme].symSort.nAttrbs;
    }

  try
    {
      ldgALhead.resize (ldgALheadIndSeq + 1, ldgAdjListHeads ());
    }
  catch (std::bad_alloc const&)
    {
      // resize of the local dependency graph adj list head failed
      yox_alloc_fail (__FILE__, __LINE__);
    }

  // clear out any existing entries
  for (ldgAdjListHeads& entry : ldgALhead)
    entry = ldgAdjListHeads ();

  ldgALheadInd[prodBufferPointer2] = ldgALheadIndSeq;
}



// handle one dependency specification (done once for each @i, @e, or @m).
void
processAdepSpec ()
{
  if (dependentSeq == 0)
    return;

  ldgAdjListHeads& pHead =
      ldgALhead[ldgALheadInd[dependent[0].pos] + dependent[0].num];

  pHead.attrbRefText = rsBuffer.get ();
  rsBuffer.clear (); // accumulate text upto the next attribute (or traversal)
                     // definition; will be ignored

  pHead.hasYaccPvarRef = hasYaccPvarRefGlob;

  for (int i = 0; i < dependentSeq; i++)
    if (dependent[i].pos != 0)
      pHead.inhFellowEnt = true;
    else
      pHead.synFellowEnt = true;

  for (int i = 0; i < dependentSeq; i++)
    {
      ldgAdjListHeads& pHead =
          ldgALhead[ldgALheadInd[dependent[i].pos] + dependent[i].num];

      if ((pHead.eeCount = dependeeSeq) == 0)
        ldgTopSortQ.push (dependent[i]);

      for (int j = 0; j < dependeeSeq; j++)
        try
          {
            pHead.eeList.push_front (dependee[j]);
          }
        catch (...)
          {
            // expansion of the dependee list failed
            yox_alloc_fail (__FILE__, __LINE__);
          }
    }

  for (int i = 0; i < dependeeSeq; i++)
    {
      attrbPairList& lp =
          ldgALhead[ldgALheadInd[dependee[i].pos] + dependee[i].num].entList;
      attrbPairList::const_iterator itr = lp.cbegin ();

      while ((itr != lp.cend ()) && (itr->pos == 0))
        ++itr;

      for (int j = 0; j < dependentSeq; j++)
        try
          {
            // inserts before 'itr', 'itr'-> inserted value
            itr = lp.insert (itr, dependent[j]);
          }
        catch (...)
          {
            // expansion of the dependent list failed
            yox_alloc_fail (__FILE__, __LINE__);
          }
    }

  dependentSeq = 0;
  dependeeSeq = 0;
  hasYaccPvarRefGlob = false;
}



static std::ostringstream formatted_msg;



static void
genAutoDef (const AGC::location& loc,
            attrbPair attrbOccur,
            bool warning,
            const std::string& fileName)
{
  attrbPair tempOccur = { 0, 0 };
  int nFound = 0;

  prodSymbol& prodSym = prodBuffer2[attrbOccur.pos];
  std::string& symName = prodSym.lexeme;
  const int& symOccr = prodSym.symOccr;

  sort& symSort = gSymTable.at (symName).symSort;
  sortAttrbSeq& sortSeq = *symSort.attrbSeq;

  std::string cp = sortSeq[attrbOccur.num].id;

  if (attrbOccur.pos == 0)
    {
      for (int i = 1; i < prodBufferPointer2; i++)
        {
          // check the RHS symbol's attributes for a match
          sort& rhsSort = gSymTable.at (prodBuffer2[i].lexeme).symSort;
          sortAttrbSeq& rhsSortSeq = *rhsSort.attrbSeq;

          const int& nA = rhsSort.nAttrbs;

          for (int j = 0; j < nA; j++)
            if (rhsSortSeq[j].id == cp)
              {
                nFound++;
                tempOccur.pos = i;
                tempOccur.num = j;
              }
        }
    }
  else
    {
      // check the LHS symbol's attributes for a match
      sort& lhsSort = gSymTable.at (prodBuffer2[0].lexeme).symSort;
      sortAttrbSeq& lhsSortSeq = *lhsSort.attrbSeq;

      const int& nA = lhsSort.nAttrbs;

      for (int j = 0; j < nA; j++)
        if (lhsSortSeq[j].id == cp)
          {
            nFound++;
            tempOccur.pos = 0;
            tempOccur.num = j;
          }
    }

  if (nFound != 1)
    {
      // error if not exactly one match

      formatted_msg << "undefined "
                       "\"" +
                           symName + "." +
                           std::to_string (symOccr) + "." +
                           cp +
                       "\"; ";

      if (nFound == 0)
        formatted_msg << "no applicable ";
      else
        formatted_msg << "too many applicable ";

      if (attrbOccur.pos == 0)
        formatted_msg << "@autosyn ";
      else
        formatted_msg << "@autoinh ";

      if (nFound == 0)
        formatted_msg << "attribute ";
      else
        formatted_msg << "attributes ";

      formatted_msg << "(line " << sortSeq[attrbOccur.num].autoDef.lineno
                    << ").\n";

      yox_error (loc, formatted_msg);
      return;
    }

  if (warning)
    {
      formatted_msg << "\"" +
                           symName + "." +
                           std::to_string (symOccr) + "." +
                           cp +
                       "\" defined by ";

      if (attrbOccur.pos == 0)
        formatted_msg << "@autosyn ";
      else
        formatted_msg << "@autoinh ";

      formatted_msg << "(line " << sortSeq[attrbOccur.num].autoDef.lineno
                    << ").\n";

      yox_warning (loc, formatted_msg);
    }

  // @autoinh or @autosyn doesn't have an input file location
  rsBuffer.init (AGC::position (new std::string (fileName), -1, 0));

  rsBuffer.putAttrbPair (attrbOccur);

  rsBuffer.gen ().put (" = ");

  memberDepend (attrbOccur, true, true);

  rsBuffer.putAttrbPair (tempOccur);

  rsBuffer.gen ().put (";\n");

  memberDepend (tempOccur, false, true);

  processAdepSpec ();

  // reset flag for processing the next rule
  markOrCheckAttrbOccrDefd (attrbOccur, true, false);
}



static bool startSymMarkedSyn = false;



// consistency and completeness check for one rule

// The calls markOrCheckAttrbOccrDefd (i, j, true, false) read whether instances
// are defined in the present attribute reference section, then reset the data
// structures they have read so that the data structures are ready for
// processing the next attribute reference section.  Each call tests for
// presence of an explicit definition (as opposed to a copy rule generated
// automatically with @autosyn or @autoinh).
static void
checkAttrbsSynOrInh (const AGC::location& loc, fileMgr& whereTo)
{
  for (int prodPos = 0; prodPos < prodBufferPointer2; prodPos++)
    {
      prodSymbol& prodSym = prodBuffer2[prodPos];
      const std::string& symName = prodSym.lexeme;
      const int& symOccr = prodSym.symOccr;

      gSym& symbol = gSymTable[symName];
      const sort& symSort = symbol.symSort;
      const sortAttrbSeq& sortSeq = *symSort.attrbSeq;

      const int& nA = symSort.nAttrbs;

      if ((prodPos == 0 /*LHS*/) && !startSymMarkedSyn &&
          (symName == startSymLexeme))
        {
          symbol.posOfDefn = loc.begin;

          for (int j = 0; j < nA; j++)
            {
              const sortAttrb& sortAttrb = sortSeq[j];
              const attrbAnnotation& autoDef = sortAttrb.autoDef;

              if (autoDef.direction == inherited)
                yox_error (loc,
                           "@autoinh "
                           "(line " + std::to_string (autoDef.lineno) + ")"
                           " for start symbol "
                           "\"" + symName + "\""
                           " attribute "
                           "\"" + sortAttrb.id + "\".\n");
              symbol.attrbDirections[j] = synthesized;
            }
          startSymMarkedSyn = true;
        }

      if (symbol.posOfDefn.line == -1)
        {
          symbol.posOfDefn = loc.begin;

          for (int j = 0; j < nA; j++)
            {
              attrbDirection& direction = symbol.attrbDirections[j];

              if (direction == undefined)
                {
                  attrbPair attrbOccur = { prodPos, j };
                  bool defdInThisProd =
                      markOrCheckAttrbOccrDefd (attrbOccur, false, false);

                  if ((defdInThisProd && (prodPos != 0 /*RHS*/)) ||
                      (!defdInThisProd && (prodPos == 0 /*LHS*/)))
                    direction = inherited;
                  else
                    direction = synthesized;
                }
            }
        }

      for (int j = 0; j < nA; j++)
        {
          attrbPair attrbOccur = { prodPos, j };

          bool defdInThisProd =
              markOrCheckAttrbOccrDefd (attrbOccur, true, false);

          attrbDirection& direction = symbol.attrbDirections[j];

          const sortAttrb& sortAttrb = sortSeq[j];
          const attrbAnnotation& autoDef = sortAttrb.autoDef;

          if (defdInThisProd)
            {
              if ((prodPos == 0 /*LHS*/) && (direction == inherited))
                {
                  if (sortSeq[j].autoDef.direction == inherited)
                    {
                      yox_error (loc,
                                 "synthesized "
                                 "\"" +
                                     symName + "." +
                                     std::to_string (symOccr) + "." +
                                     sortAttrb.id +
                                 "\""
                                 " is declared @autoinh "
                                 "(line " +
                                     std::to_string (autoDef.lineno) + ").\n");
                    }
                  else
                    {
                      formatted_msg << "synthesized "
                                       "\"" + symName + "." +
                                           std::to_string (symOccr) + "." +
                                           sortAttrb.id +
                                       "\""
                                       " is inherited (line ";
                      if (*symbol.posOfDefn.filename != *loc.begin.filename)
                        formatted_msg << *symbol.posOfDefn.filename + ":";
                      formatted_msg << symbol.posOfDefn.line << ").\n";
                      yox_error (loc, formatted_msg);
                    }
                }
              if ((prodPos != 0 /*RHS*/) && (direction == synthesized))
                {
                  if (sortSeq[j].autoDef.direction == synthesized)
                    {
                      yox_error (loc,
                                 "inherited "
                                 "\"" +
                                     symName + "." +
                                     std::to_string (symOccr) + "." +
                                     sortAttrb.id +
                                 "\""
                                 " is declared @autosyn "
                                 "(line " +
                                     std::to_string (autoDef.lineno) + ").\n");
                    }
                  else
                    {
                      formatted_msg << "inherited "
                                       "\"" +
                                           symName + "." +
                                           std::to_string (symOccr) + "." +
                                           sortAttrb.id +
                                       "\""
                                       " is synthesized (line ";
                      if (*symbol.posOfDefn.filename != *loc.begin.filename)
                        formatted_msg << *symbol.posOfDefn.filename + ":";
                      formatted_msg << symbol.posOfDefn.line << ").\n";
                      yox_error (loc, formatted_msg);
                    }
                }
            }
          else
            {
              std::string fileName = whereTo.getName ();
              if ((prodPos == 0 /*LHS*/) && (direction == synthesized))
                {
                  if (autoDef.direction == synthesized)
                    genAutoDef (loc, attrbOccur, autoDef.autoWarn, fileName);
                  else
                    {
                      formatted_msg << "undefined "
                                       "\"" +
                                           symName + "." +
                                           std::to_string (symOccr) + "." +
                                           sortAttrb.id +
                                       "\""
                                       " is synthesized (line ";
                      if (*symbol.posOfDefn.filename != *loc.begin.filename)
                        formatted_msg << *symbol.posOfDefn.filename + ":";
                      formatted_msg << symbol.posOfDefn.line << ").\n";
                      yox_error (loc, formatted_msg);
                    }
                }
              if ((prodPos != 0 /*RHS*/) && (direction == inherited))
                {
                  if (autoDef.direction == inherited)
                    genAutoDef (loc, attrbOccur, autoDef.autoWarn, fileName);
                  else
                    {
                      formatted_msg << "undefined "
                                       "\"" +
                                           symName + "." +
                                           std::to_string (symOccr) + "." +
                                           sortAttrb.id +
                                       "\""
                                       " is inherited (line ";
                      if (*symbol.posOfDefn.filename != *loc.begin.filename)
                        formatted_msg << *symbol.posOfDefn.filename + ":";
                      formatted_msg << symbol.posOfDefn.line << ").\n";
                      yox_error (loc, formatted_msg);
                    }
                }
            } // else

        } // for j

    } // for i
}



// put synthesized attributes of tokens in the ready set (queue)
static void
enqueueSynAttrbsOfTerms (void)
{
  for (int i = 1; i < prodBufferPointer2; i++)
    {
      gSym& symbol = gSymTable[prodBuffer2[i].lexeme];

      for (int j = 0; j < symbol.symSort.nAttrbs; j++)
        {
          if ((symbol.attrbDirections[j] == synthesized) && (symbol.symNum > 0))
            ldgTopSortQ.push (attrbPair (i, j));
        }

    } // for i
}



// Set to true in 'ldgTopSort ()' if there are any `yyyDecorate()' evaluations
bool genRSNodeProdNumSwitchHd = false;

// the static part of the topological sort.
// outputs semantic actions into the Yacc actions or the big nested
// switch statement in the `SolveAndSignal()' portion of `yyyDecorate()'.
static void
ldgTopSort (const AGC::location& loc)
{
  const yoxControl& ctrl = outMgr::ctrl;

  static long entOccrSeqLoc;

  long entOccrSeqLocHold;

//std::cerr << prodNumSeq -1 << ": \n";

  enqueueSynAttrbsOfTerms ();

  while (!ldgTopSortQ.empty ())
    {
      attrbPair aoDum = ldgTopSortQ.front ();
      ldgTopSortQ.pop ();

      struct ldgAdjListHeads& pHead =
          ldgALhead[ldgALheadInd[aoDum.pos] + aoDum.num];

//std::cerr << "\n    " +
//              prodBuffer2 [aoDum.pos].lexeme << "." +
//              gSymTable[prodBuffer2[aoDum.pos].lexeme].sortSeq[aoDum.num].id +
//              "\n";

      if (!pHead.attrbRefText.buffer.empty ())
        {
          yoxOutG.translateAndOutput (imm,
                                      pHead.attrbRefText,
                                      attrbPair (0 /*don't care*/,
                                                 0 /*don't care*/));
          if (ctrl.noLineDirectives)
            yoxOutG.put ("\n");
        }

      if (!pHead.entList.empty ())
        for (const attrbPair& itr : pHead.entList)
          {
            if (--(ldgALhead[ldgALheadInd[itr.pos] + itr.num].eeCount) == 0)
              ldgTopSortQ.push (itr);
          }
    }

  // separate the part above this comment and the part below into two separate
  // routines.  The above is truly the sort.

  bool genWhichSymSwitchHd = false;
  for (int i = 0; i < prodBufferPointer2; i++)
    {
      bool genWhichAttrbSwitchHd = false;

      const int& nA = gSymTable[prodBuffer2[i].lexeme].symSort.nAttrbs;

      if (nA > 0)
        {
          if (!genRSNodeProdNumSwitchHd)
            {
              genRSNodeProdNumSwitchHd = true;
              temp1Out.putSolveAndSignalHeader ();
            }

          if (!genWhichSymSwitchHd)
            {
              genWhichSymSwitchHd = true;

              const std::string ind = std::to_string (prodNumSeq - 1);
              temp1Out.gen ()
                  .put ("case " + ind + ":" "  /***Yacc rule " + ind + "***/\n"
                        "  switch (whichSym) {\n");
            }

          for (int j = 0; j < nA; j++)
            {
              struct ldgAdjListHeads& pHead = ldgALhead[ldgALheadInd[i] + j];

              entOccrSeqLocHold = entOccrSeqLoc;

              if (pHead.eeCount != 0)
                {
                  if (!pHead.entList.empty ())
                    for (const attrbPair& itr __attribute__ ((unused)) :
                         pHead.entList)
                      entOccrSeqLoc++;
                }

              if (pHead.eeCount > 0)
                {
                  temp1Out.gen ();

                  if (!genWhichAttrbSwitchHd)
                    {
                      genWhichAttrbSwitchHd = true;

                      temp1Out.put ("  case " + std::to_string (i) + ":  /**/\n"
                                    "    switch (whichAttrb) {\n");
                    }

                  temp1Out.put ("    case " + std::to_string (j) + ":\n");

                  if (!pHead.attrbRefText.buffer.empty ())
                    {
                      temp1Out.translateAndOutput (decor,
                                                   pHead.attrbRefText,
                                                   attrbPair (i, j));
                      if (ctrl.noLineDirectives)
                        temp1Out.put ("\n");
                    }

                  if (entOccrSeqLocHold < entOccrSeqLoc)
                    {
                      temp1Out.gen ()
                          .put (yoxPrefix ("SignalEnts") + "(RSNode," +
                                std::to_string (entOccrSeqLocHold * 2) + "," +
                                std::to_string (entOccrSeqLoc * 2) + ");\n");
                    }

                  if (pHead.hasYaccPvarRef)
                    yox_error (loc,
                               "decoration-time reference to "
                               "Yacc pseudo variable.\n");

                  if (pHead.inhFellowEnt && pHead.synFellowEnt)
                    yox_error (loc,
                               "attribute reference segment with "
                               "inherited & synthesized dependents.\n");

                  temp1Out.gen ()
                      .put ("    break;\n");
                }

              pHead.synFellowEnt = false;
              pHead.inhFellowEnt = false;
              pHead.hasYaccPvarRef = false;

            } // for j

          if (genWhichAttrbSwitchHd)
            temp1Out.gen ()
                .put ("    }\n"
                      "  break;\n");
        }

    } // for i

  if (genWhichSymSwitchHd)
    temp1Out.gen ()
        .put ("  }\n"
              "break;\n");
}



static void
printProd (void)
{
  if ((((prodNumSeq - 1) - 1) % 10) == 0)
    tempProdsIndFile.put ("\n");

  tempProdsIndFile.put (gsOccurSeq, 4).put (",");

  for (int k = 0; k < prodBufferPointer2; k++)
    {
      const std::string& symbol = prodBuffer2[k].lexeme;

      if (((gsOccurSeq + k) % 5) == 0)
        tempProdsNumsFile.put ("\n");

      tempProdsNumsFile
          .put ("{")
          .put (stringSerialNoLookup (symbol), 4)
          .put (",")
          .put (gSymTable[symbol].symSort.typeNum, 4)
          .put ("},");
    }

  gsOccurSeq += prodBufferPointer2;
}


#if defined(G_DEBUG)
void
showDepGraph (void)
{
  std::cerr << "***************  dependency graph for production "
            << prodNumSeq - 1 << ":";

  for (int i = 0; i < prodBufferPointer2; i++)
    {
      const int& nA = gSymTable[prodBuffer2[i].lexeme].symSort.nAttrbs;

      for (int j = 0; j < nA; j++)
        {
          std::cerr << "\n\ndependees  of " << i << " " << j << ":  ";

          attrbPairList& nextANee = ldgALhead[ldgALheadInd[i] + j].eeList;

          if (!nextANee.empty ())
            for (const attrbPair& itr : nextANee)
              std::cerr << "    " << itr.pos << " " << itr.num;

          std::cerr << "\ndependents of " << i << " " << j << ":  ";

          attrbPairList& nextANent = ldgALhead[ldgALheadInd[i] + j].entList;
          if (!nextANent.empty ())
            for (const attrbPair& itr : nextANent)
              std::cerr << "    " << itr.pos << " " << itr.num;

          std::cerr << "\n";

        } // for j

    } // for i

  std::cerr << "\n\n\n";
}



// for debugging the grammar-symbol table
void
showGsymInfo (int symPos)
{
  std::string& symbol = prodBuffer2[symPos].lexeme;

  gSym& entry = gSymTable[prodBuffer2[symPos].lexeme];
  sort& symSort = entry.symSort;
  sortAttrbSeq& sortSeq = *symSort.attrbSeq;

  std::cerr << "     symbol Number: " << entry.symNum << "\n"
            << "     type :         " << symSort.typeNum << "\n"
            << "     lexeme:        " << symbol << "\n"
            << "     number of attributes: " << symSort.nAttrbs << "\n";

  for (int j = 0; j < symSort.nAttrbs; j++)
    {
      std::cerr << "          attribute " << std::setw (2) << j << ": "
                << sortSeq[j].id;

      if (entry.attrbDirections[j] != undefined)
        {
          std::cerr << "  defined at line: " << entry.posOfDefn.line
                    << " ";
          if (entry.attrbDirections[j] == inherited)
            std::cerr << "  inherited\n";
          else
            std::cerr << "  synthesized\n";
        }
      else
        {
          std::cerr << " *neither syn nor inh \n";
        }
    }
}



void
showSymInfoForProd (void)
{
  std::cerr << "\n"
            << "Symbol information for production " << prodNumSeq - 1
            << " : \n";

  for (int k = 0; k < prodBufferPointer2; k++)
    {
      std::cerr << "Position: " << k << "\n";
      showGsymInfo (k);
    }

  std::cerr << "\n\n\n";
}



void
othershowGsymInfo (void)
{
  std::cerr << "+++++++++++++++++symbol info for production " << prodNumSeq - 1
            << ":\n";

  for (int i = 0; i < prodBufferPointer2; i++)
    {
      std::string symbol = prodBuffer2[i].lexeme;

      gSym& entry = gSymTable[symbol];
      sort& symSort = entry.symSort;
      sortAttrbSeq& sortSeq = *symSort.attrbSeq;

      const int& nA = symSort.nAttrbs;

      for (int j = 0; j < nA; j++)
        {
          std::cerr << symbol << "  " << sortSeq[j].id
                    << "  defined at line: " << entry.posOfDefn.line
                    << ".";
          if (entry.attrbDirections[j] == inherited)
            std::cerr << " inherited\n";
          else if (entry.attrbDirections[j] == synthesized)
            std::cerr << " synthesized\n";
          else
            std::cerr << " undefined\n";

        } // for j

    } // for i

  std::cerr << "\n\n\n\n";
}
#endif



static int ldgCDcount;



// Returns false if pred is true for some attribute occurrence
// in the graph closure of AO.
//
// The graph is 'dependee'.
//
// Returns true if pred is false for every attribute occurrence
// in the graph closure of AO.
//
// In other words, returns true exactly when AO's graph closure
// is disjoint from the set of occurrences that satisfy pred.
//
// Used for computing pruning conditions.
static bool
ldgClosureDisjoint (const AGC::location& loc,
                    const attrbPair& AO,
                    bool (*pred) (attrbPair))
{
  if (ldgCDcount++ > 200)
    {
      yox_warning (loc, "rule probably has cycle.\n");
      return false;
    }

  if ((*pred) (AO))
    return false;

  attrbPairList& nextAN = ldgALhead[ldgALheadInd[AO.pos] + AO.num].eeList;

  if (!nextAN.empty ())
    for (const attrbPair& itr : nextAN)
      if (!ldgClosureDisjoint (loc, itr, pred))
        return false;

  return true;
}



static bool
isLHSinh (attrbPair AO)
{
  return ((AO.pos == 0) &&
          (gSymTable[prodBuffer2[AO.pos].lexeme].attrbDirections[AO.num] ==
           inherited));
}



static bool
areAllRHSoccrsSolvable (const AGC::location& loc)
{
  ldgCDcount = 0;
  for (int i = 1; i < prodBufferPointer2; i++)
    {
      attrbPair AO;
      AO.pos = i;
      const int& nA = gSymTable[prodBuffer2[i].lexeme].symSort.nAttrbs;

      for (int j = 0; j < nA; j++)
        {
          AO.num = j;
          if (!ldgClosureDisjoint (loc, AO, isLHSinh))
            return false;
        }

    } // for i

  return true;
}



static bool
onRHS (attrbPair AO)
{
  return (AO.pos > 0);
}



static bool
oneSynEELevel (const AGC::location& loc)
{
  ldgCDcount = 0;
  gSym& symbol = gSymTable.at (prodBuffer2[0].lexeme);
  const int& nA = symbol.symSort.nAttrbs;
  attrbPair AO;
  AO.pos = 0;

  for (int j = 0; j < nA; j++)
    {
      AO.num = j;
      if (symbol.attrbDirections[j] == synthesized)
        if (!ldgClosureDisjoint (loc, AO, isLHSinh) &&
            !ldgClosureDisjoint (loc, AO, onRHS))
          return false;
    }

  return true;
}



static bool allRHSoccrsSolvable, singleLevelSynEE;

static void
doPruneCondChecks (const AGC::location& loc)
{
  allRHSoccrsSolvable = areAllRHSoccrsSolvable (loc);
  singleLevelSynEE = oneSynEELevel (loc);
}



// output the dependee-count initialization lists and indices into
// the space of those lists
static void
conveyLocalDepGraphInfo (void)
{
  const yoxControl& ctrl = outMgr::ctrl;

  static int k1 = 0, k2 = 0;

  static long GSoccrSeq;

//static long entOccrSeqHold;

  long eeCountInitCurrentProd = eeCountInitSeq;

  indToEntsByProdNum[prodNumSeq - 1] = GSoccrSeq;

  std::vector<ldgAdjListHeads>::iterator pHead = ldgALhead.begin ();

  for (int i = 0; i < prodBufferPointer2; i++)
    {
      temp5Out.put (attrbOccrSeq).put (",");

      if ((++k1 % 12) == 0)
        temp5Out.put ("\n");

      GSoccrSeq++;
      const int nA = (int)(ldgALheadInd[i + 1] - ldgALheadInd[i]);
      for (int j = 0; j < nA; j++)
        {
          if (pHead->eeCount != -1)
            {
              if (i == 0)
                {
                  // LHS
                  if (permitCtimeCircTest)
                    {
                      temp2Out
                          .put (yoxPrefix ("R") + ",")
                          .put (j).put (",")
                          .put (pHead->eeCount).put (", ");

                      eeCountInitSeq++;
                    }
                }
              else
                {
                  // RHS
                  temp2Out
                      .put (i - 1).put (",")
                      .put (j).put (",")
                      .put (pHead->eeCount).put (", ");

                  eeCountInitSeq++;
                }

              if ((eeCountInitSeq % 6) == 0)
                temp2Out.put ("\n");

              pHead->eeList.clear ();
            }

//        entOccrSeqHold = entOccrSeq;

          temp4Out.put (entOccrSeq * 2).put (",");

          if ((++k2 % 8) == 0)
            temp4Out.put ("\n");

          attrbOccrSeq++;
          if (pHead->eeCount != 0)
            {
              attrbPairList& nextAN = pHead->entList;
              if (!nextAN.empty ())
                for (const attrbPair& itr : nextAN)
                  {
                    temp3Out.put (itr.pos).put (",").put (itr.num).put (",");

                    if (((++entOccrSeq) % 8) == 0)
                      temp3Out.put ("\n");
                  }
            }

//        temp6Out contents never used
//        temp6Out
//            .put ((unsigned long) (entOccrSeq - entOccrSeqHold)).put (",");
//        if (!(k2 % 8))
//          temp6Out.put ("\n");

          pHead->eeCount = -1;
          pHead->entList.clear ();
          ++pHead;

        } // for j

    } // for i

  yoxOutG.gen ()
      .put (yoxPrefix ("AdjustINRC") + "(" +
            std::to_string (prodNumSeq - 1) + "," +
            std::to_string (prodBufferPointer2 - 1) + "," +
            std::to_string (eeCountInitCurrentProd * 3) + "," +
            std::to_string (eeCountInitSeq * 3) + ",")
      .putLHSsemStkElRef ()
      .putRHSsemStkElRefList ()
      .put (");\n");

  bool prune = false;
  if ((prodBufferPointer2 > 1) && allRHSoccrsSolvable && singleLevelSynEE)
    {
      if (!ctrl.suppressPruning)
        {
          yoxOutG
              .putDecorateCall ("true") // -> pruning
              .put (yoxPrefix ("Prune") + "(" +
                    std::to_string (prodNumSeq - 1) + ",")
              .putLHSsemStkElRef ()
              .put (");\n");

          prune = true;
        }
    }
  if (!prune)
    {
      if (decorationRequested)
        yoxOutG.putDecorateCall ("true"); // -> pruning
      else
        yoxOutG.putDecorateCall ("false"); // -> !pruning
    }
}



// writes the list of solved synthesized attributes of each token
const refCount
installLeafRefCountInitList (const std::string& token)
{
  long leafRCIShold;

  if (!permitCtimeCircTest)
    return refCount (); // dummy value

  leafRCIShold = leafRefCountInitSeq;

  gSym& symbol = gSymTable[token];

  if (symbol.symNum < 0)
    {
      // returning nonterminal as token: error previously reported
      return refCount (); // dummy value
    }

  if (((long)leafRefCount.size ()) - 1 < symbol.symNum)
    try
      {
        leafRefCount.resize (symbol.symNum + NUMTERMSINCR, refCount ());
      }
    catch (std::bad_alloc const&)
      {
        // reallocation of leaf ref count data list stg failed
        yox_alloc_fail (__FILE__, __LINE__);
      }

  if (leafRefCount[symbol.symNum].start == -1)
    {
      leafRefCount[symbol.symNum].start = leafRefCountInitSeq;

      for (int i = 0; i < symbol.symSort.nAttrbs; i++)
        {
          if (symbol.attrbDirections[i] == synthesized)
            {
              temp7Out.put (i).put (",");

              if (((++leafRefCountInitSeq) % 20) == 0)
                temp7Out.put ("\n");
            }
        }

      leafRefCount[symbol.symNum].length =
          leafRefCountInitSeq - leafRCIShold;
    }

  return leafRefCount[symbol.symNum];
}



void
finishYfileAttrbRefSection (const AGC::location& loc)
{
#if defined(G_DEBUG)
  showDepGraph ();
#endif

  checkAttrbsSynOrInh (loc, yoxOutG);

#if defined(G_DEBUG)
  showDepGraph ();
#endif

  doPruneCondChecks (loc);

#if defined(G_DEBUG)
  showSymInfoForProd ();
#endif

  printProd ();

  yoxOutG.gen ()
      .put ("\n{if(" + yoxPrefix ("Yok") + "){\n")
      .putGenIntNodeCall ();

  ldgTopSort (loc);

  tempTravOut.putTravCodePerProd ();

  conveyLocalDepGraphInfo ();

  yoxOutG
      .putReclaimSemStkEls () // reclaim the RHS semantic stack element(s)
      .put ("}}\n");

  // dump the closing '}' of the Yacc action, if any
  yoxOutG.append (actionCLOSER);
  actionCLOSER.clear ();
}



void
finishLfileAttrbRefSection (const AGC::location& loc)
{
  checkAttrbsSynOrInh (loc, yoxOutL);
}
