/* $Id: fileMgr.cc 2838 2022-12-11 23:01:52Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2010-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <cassert>

#include <string>

#include <fstream>
#include <sstream>

#include <cerrno>
#include <cstring> // for 'strerror ()'

#include <system_error>

#include "fileMgr.hh"
#include "memMgr.hh"
#include "messages.hh"
#include "outMgr.hh"



bool
erase (const std::string& fileName)
{
  return (std::remove (fileName.c_str ()) == 0);
}


// =============================== class fileMgr ===============================

// public:

fileMgr::fileMgr (std::string fileName, bool forLexer)
    : codeGen (new std::fstream (fileName, std::ios_base::out),
               fileName,
               -1,
               nullptr,
               forLexer)
{
  if (!getObj ().is_open ())
    {
      const AGC::location loc; // dummy location
      yox_fatal (loc, "opening " + fileName + " failed.\n");
    }

  wasOutName = true;

  // just in case this is a temporary output file
  origin = AGC::position (&name);
  curPos = &origin;
}



void
fileMgr::open (std::string fileName)
{
  getObj ().open (fileName, std::ios_base::out);
  if (!getObj ().is_open ())
    {
      const AGC::location loc; // dummy location
      yox_fatal (loc, "opening " + fileName + " failed.\n");
    }

  clear ();

  name = fileName;

  wasOutName = true;

  // just in case this is a temporary output file
  origin = AGC::position (&name);
  curPos = &origin;
}



void
fileMgr::close (void)
{
  if (!getObj ().is_open ())
    return;

  flush (); // write any currently held & buffered text to the output object

  getObj ().close ();
}



void
fileMgr::erase (void)
{
  if (!::erase (name))
    {
      const AGC::location loc; // dummy location
      yox_fatal (loc,
                 "removing " + name + ": system error: " +
                     strerror (errno) + ".\n");
    }
}



fileMgr&
fileMgr::put (char c)
{
  putCharInWriteBuffer (c);

  return *this;
}



// Flushes & rewinds the open whereFrom fileMgr, then copies it to this
// fileMgr already open for writing. Convert #line directives that
// need to refer to the destination filename.
fileMgr&
fileMgr::append (fileMgr& whereFrom)
{
  assert (whereFrom != *this);

  if (whereFrom.empty ())
    return *this;

  std::fstream& from = whereFrom.getObj ();

  // close & reopen the external file for input
  from.close ();
  from.open (whereFrom.name, std::ios_base::in);

  outMgr::append (from, whereFrom.name);

  return *this;
}



fileMgr&
fileMgr::append (lookaheadBuf& whereFrom)
{
  assert (whereFrom != *this);

  if (whereFrom.empty ())
    return *this;

  if (!ctrl.noLineDirectives)
    putInLineDirective (whereFrom.origin.line, *whereFrom.origin.filename);

  std::stringstream& from = whereFrom.getObj ();

  // rewind the internal memory file
  from.clear (); // clear the stream error flags
  from.seekg (0); // ... and then rewind it

  outMgr::append (from, whereFrom.name);

  return *this;
}


// protected:

// private:

// ================================ end fileMgr ================================
