/* $Id: YY_LOCATIONS.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2019-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef YY_LOCATIONS_HH_INCLUDED
#define YY_LOCATIONS_HH_INCLUDED

// customize location processing

// custom definition of 'struct yy_buffer_state':
// preserves column count, and the EOL context used for 'yyless(n)'
// push back (only works for push back of at most one trailing EOL)
struct yy_buffer_state
{
  std::streambuf* yy_input_file;

  char* yy_ch_buf;  // input buffer
  char* yy_buf_pos; // current position in input buffer

  // Size of input buffer in bytes, not including room for EOB characters.
  int yy_buf_size;

  // Number of characters read into yy_ch_buf, not including EOB characters.
  int yy_n_chars;

  // Whether we "own" the buffer - i.e., we know we created it,
  // and can realloc() it to grow it, and should free() it to
  // delete it.
  int yy_is_our_buffer;

  // Whether this is an "interactive" input source; if so, and
  // if we're using stdio for input, then we want to use getc()
  // instead of fread(), to make sure we stop fetching input after
  // each newline.
  int yy_is_interactive;

  // Whether we're considered to be at the beginning of a line.
  // If so, '^' rules will be active on the next match, otherwise not.
  int yy_at_bol;

  int yy_bs_lineno;        // The line count
  int yy_bs_column;        // The column count.
  AGC::location yy_bs_loc; // The current location

  EOLctx yy_bs_EOLstate;

  // Whether to try to fill the input buffer when we reach the end of it.
  int yy_fill_buffer;

  int yy_buffer_status;

#define YY_BUFFER_NEW 0
#define YY_BUFFER_NORMAL 1
#define YY_BUFFER_EOF_PENDING 2
  // When an EOF's been seen but there's still some text to process
  // then we mark the buffer as YY_EOF_PENDING, to indicate that we
  // shouldn't try reading from the input source any more.  We might
  // still have a bunch of tokens to match, though, because of
  // possible backing-up.

  // When we actually see the EOF, we change the status to "new"
  // (via yyrestart()), so that the user can continue scanning by
  // just pointing yyin at a new input file.
};



// custom definition of 'YY_LESS_LINENO(n)' macro:
// Ox doesn't back up over an EOL
#undef YY_LESS_LINENO
#define YY_LESS_LINENO(n) \
{ \
  char* pos = std::strrchr (yytext, '\t'); \
  if (pos == nullptr || (pos - yytext) < n) \
    yycolumn -= (yyleng - n); \
  else if (n < yyleng) \
    { \
      for (int i = yyleng - 1; i >= n; i--) \
        { \
          yycolumn--; \
          if (yytext[i] == '\t') \
            yycolumn -= (7 - yycolumn % 8); \
        } \
    } \
  yylloc.end.line = yylineno; \
  yylloc.end.column = yycolumn; \
}



// Ox only uses 'yyless (n)'; alert use of 'unput()'
#undef unput



#define YY_USER_ACTION \
{ \
  if (EOLstate.atEOL) \
    { \
      EOLstate.atEOL = false; \
      yylineno = EOLstate.EOLpos.line + 1; \
      yycolumn = 1; \
    } \
  yylloc.begin.line = yylineno; \
  yylloc.begin.column = yycolumn; \
  if (std::strchr (yytext, '\t') == nullptr) \
    yycolumn += yyleng; \
  else \
    for (int i = 0; i < yyleng; i++) \
      { \
        if (yytext[i] == '\t') \
          yycolumn += (8 - yycolumn % 8); \
        yycolumn++; \
      } \
  yylloc.end.line = yylineno; \
  yylloc.end.column = yycolumn; \
  if (yytext[yyleng - 1] == '\n') \
    { \
      EOLstate.atEOL = true; \
      /* save the EOL posiiton */ \
      EOLstate.EOLpos = position (nullptr, yylineno, yycolumn); \
      yylloc.end.line++; \
      yylloc.end.column = 0; \
    } \
}

#endif // !YY_LOCATIONS_HH_INCLUDED
