/* $Id: agc-strip-lexer.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef AGC_STRIP_LEXER_HH_INCLUDED
#define AGC_STRIP_LEXER_HH_INCLUDED

#if !defined(yyFlexLexer)
#define yyFlexLexer strip_FlexLexer
#include <FlexLexer.h>

// 'flex' secrets:
#define BEGIN (yy_start) = 1 + 2 *
#define YY_START (((yy_start) - 1) / 2)
#endif

#include <fstream>
#include <stack>

namespace AGC
{
namespace STRIP
{

class Lexer : public strip_FlexLexer
{
public:

  Lexer (specIO&, std::ifstream&);

  inline int
  yylex (lexer::value_type* yylval, location* yylloc)
  {
    return yylex (*yylval, *yylloc);
  }

  int yylex (lexer::value_type& yylval, location& yylloc);

  inline int
  yylex (void) // should never be called
  {
    return lexer::symbol_kind::S_YYerror;
  }

  inline int&
  lineno (void)
  {
    return yylineno;
  }

  inline int&
  columno (void)
  {
    return yycolumn;
  }

protected:

  inline int
  start (void) const
  {
    return YY_START;
  }

  inline void
  start (int new_state)
  {
    BEGIN (new_state);
  }

  inline size_t
  size (void) const
  {
    return (size_t)yyleng;
  }

  inline int
  chr (void) const
  {
    return (int)*yytext;
  }

  inline const char*
  text (void) const
  {
    return yytext;
  }

  inline const std::string
  str (void) const
  {
    return std::string (yytext);
  }

  inline void
  push_state (const position& pos, int new_state)
  {
    yy_push_state (new_state);
    startPos.push (pos); // starting position if EOF in 'new_state'
  }

  inline void
  push_state (const location& loc, int new_state)
  {
    yy_push_state (new_state);
    startPos.push (loc.begin); // starting position if EOF 'new_state'
  }

  inline void
  pop_state (void)
  {
    yy_pop_state ();
    startPos.pop ();
  }

#if defined(_WIN32)
  int LexerInput (char*, int); // overload to treat '\r' or "\r\n" as '\n'
#endif

  specIO& theFiles;

  EOLctx EOLstate;

private:

  int yycolumn;

  int nLbraces; // '{' ... '}' nesting

  int yfileMarkSeen; // count of "%%" in the Y-file

  outMgr* whereTo;
  void
  yox_push_state (int, outMgr&);

  std::stack<position> startPos;

  bool stringHasBareNewline;
};

}
}

// begin hack: required by broken, IMHO, 'flex' C++ lexer implememtation

// There is no way to override the signature of 'yylex()' in the base
// classes 'FlexLexer' and 'yyFlexLexer'.

// A Bison C++ parser is reentrant and expects the following signature:
//     int yylex (YYSTYPE* yylval, YYLTYPE* yylloc)

// The 'YY_DECL' macro can define the signature of the 'yylex ()' member
// function in an application lexer derived from 'yyFlexLexer', leaving
// the 'yyFlexLexer::yylex (void)' member function hanging out in the cold;
// here is it:

inline int // no function actually generated
yyFlexLexer::yylex (void)
{
  return 0; // never called
}

// end hack

#endif // !AGC_STRIP_LEXER_HH_INCLUDED
