/* $Id: messages.cc 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2019-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <iomanip>
#include <ostream>
#include <sstream>

#include "location.hh"

#include "mac.hh"
#include "messages.hh"
#include "opts.hh"
#include "scan.hh"

// error/warning messages used in multiple places

void
ambiguousReturn (const AGC::location& loc)
{
  yox_warning (loc, "object of return is ambiguous.\n");
}



void
conflictingYYSTYPEdefinitions (const AGC::location& loc,
                               const std::string& current,
                               const std::string& prior)
{
  yox_warning (loc,
               "\"" + current + "\" may conflict with prior \"" + prior +
                   "\".\n");
}



void
escapeCharMisuse (const AGC::location& loc)
{
  yox_error (loc, "misuse of \'\\\' escape character in macro invocation.\n");
}



void
multipleYYSTYPEdefinitions (const AGC::location& loc, const std::string& def)
{
  yox_warning (loc, "multiple \"" + def + "\" definitions may conflict.\n");
}



void
multipleReturnTokens (const AGC::location& loc)
{
  yox_warning (loc, "object of return (multiple tokens) is ambiguous.\n");
}



void
notSupported (const AGC::location& loc, const std::string& what)
{
  yox_error (loc, what + " is not supported.\n");
}



void
returnNonterminalAsToken (const AGC::location& loc, const std::string& id)
{
  yox_error (loc, "to return nonterminal \"" + id + "\" as token.\n");
}



void
stringContainsNewline (const AGC::location& curLoc,
                       const AGC::position& startPos)
{
  AGC::location loc (startPos, curLoc.end);
  yox_error (loc, "string contains unescaped newline(s).\n");
}



void
unbalancedParensInMacro (const AGC::location& loc)
{
  yox_error (loc, "unbalanced parentheses in macro invocation.\n");
}



void
unexpectedATLBRACE (const AGC::location& loc)
{
  yox_error (loc, "unexpected \"@{\".  Missing \"@}\"?\n");
}



void
unexpectedChar (const AGC::location& loc, char what, const std::string& where)
{
  std::ostringstream message;
  message << "unexpected character \'";
  if (isprint (what))
    message << what;
  else
    message << "\\" << std::setw (3) << std::setfill ('0') << std::oct
            << (unsigned int)what;
  message << "\' " << where << ".\n";
  yox_error (loc, message);
}



void
unexpectedEOF (const AGC::location& loc)
{
  yox_fatal (loc, "unexpected end of file.\n");
}



void
unknownNodeType (const AGC::location& loc)
{
  yox_warning (loc, "unknown node type--assuming no attributes.\n");
}



void
unmatchedRPAREN (const AGC::location& loc)
{
  yox_error (loc, "unmatched \')\'.\n");
}



void
unterminatedAction (const AGC::location& curLoc, const AGC::position& startPos)
{
  AGC::location loc (startPos, curLoc.begin);
  if (curLoc.end.column != 0)
    {
      // no last EOL char
      loc.end = curLoc.end;
    }
  yox_fatal (loc, "unterminated rule action.\n");
}



void
unterminatedAttrDeclaration (const AGC::location& curLoc,
                             const AGC::position& startPos)
{
  AGC::location loc (startPos, curLoc.begin);
  if (curLoc.end.column != 0)
    {
      // no last EOL char
      loc.end = curLoc.end;
    }
  yox_fatal (loc, "unterminated attribute declaration.\n");
}



void
unterminatedAttrRefSection (const AGC::location& curLoc,
                            const AGC::position& startPos)
{
  AGC::location loc (startPos, curLoc.begin);
  if (curLoc.end.column != 0)
    {
      // no last EOL char
      loc.end = curLoc.end;
    }
  yox_fatal (loc, "unterminated attribute reference section.\n");
}



void
unterminatedCode (const AGC::location& curLoc, const AGC::position& startPos)
{
  AGC::location loc (startPos, curLoc.begin);
  if (curLoc.end.column != 0)
    {
      // no last EOL char
      loc.end = curLoc.end;
    }
  yox_fatal (loc, "unterminated C code section.\n");
}



void
unterminatedComment (const AGC::location& curLoc, const AGC::position& startPos)
{
  AGC::location loc (startPos, curLoc.begin);
  if (curLoc.end.column != 0)
    {
      // no last EOL char
      loc.end = curLoc.end;
    }
  yox_fatal (loc, "unterminated comment.\n");
}



void
unterminatedCPPdirective (const AGC::location& curLoc,
                          const AGC::position& startPos)
{
  AGC::location loc (startPos, curLoc.begin);
  if (curLoc.end.column != 0)
    {
      // no last EOL char
      loc.end = curLoc.end;
    }
  yox_fatal (loc, "unterminated C/C++ preprocessor directive.\n");
}



void
unterminatedString (const AGC::location& curLoc, const AGC::position& startPos)
{
  AGC::location loc (startPos, curLoc.begin);
  if (curLoc.end.column != 0)
    {
      // no last EOL char
      loc.end = curLoc.end;
    }
  yox_fatal (loc, "unterminated string.\n");
}



void
unterminatedYaccUnion (const AGC::location& curLoc,
                       const AGC::position& startPos)
{
  AGC::location loc (startPos, curLoc.begin);
  if (curLoc.end.column != 0)
    {
      // no last EOL char
      loc.end = curLoc.end;
    }
  yox_fatal (loc, "unterminated Yacc `%union' declaration.\n");
}



static void
yox_message (const AGC::location& loc, const char* kind, std::string msg)
{
  std::cerr << "Ox: ";
  if (expandingMacro () && replacingFormal ())
    std::cerr << formalLoc () << ": ";
  else
    std::cerr << loc << ": ";
  if (kind != nullptr)
    std::cerr << kind << ": ";
  std::cerr << msg;
  if (expandingMacro ())
    std::cerr << " [macro invoked on line " << macroInvokedFrom () << "].\n";
}



void
yox_error (const AGC::location& loc, const std::string& msg)
{
  yox_message (loc, "error", msg.c_str ());
  outMgr::ctrl.removeOutput = true; // discard generated output
}



void
yox_error (const AGC::location& loc, std::ostringstream& msg_sstr)
{
  std::string msg = msg_sstr.str ();
  msg_sstr.str (std::string ());
  msg_sstr.clear (); // reset the stringstream

  yox_message (loc, "error", msg.c_str ());
  outMgr::ctrl.removeOutput = true; // discard generated output
}



void
yox_warning (const AGC::location& loc, const std::string& msg)
{
  yox_message (loc, "warning", msg);
}



void
yox_warning (const AGC::location& loc, std::ostringstream& msg_sstr)
{
  std::string msg = msg_sstr.str ();
  msg_sstr.str (std::string ());
  msg_sstr.clear (); // reset the stringstream

  yox_message (loc, "warning", msg.c_str ());
}



void
yox_fatal (const AGC::location& loc, const std::string& msg)
{
  yox_message (loc, "fatal error", msg);

  outMgr::ctrl.removeOutput = true; // discard generated output
  exit (EXIT_FAILURE);
}



void
yox_fatal (const AGC::location& loc, std::ostringstream& msg_sstr)
{
  const std::string& msg = msg_sstr.str ();
//msg_sstr.str (std::string ());
//msg_sstr.clear (); // reset the stringstream

  yox_message (loc, "fatal error", msg.c_str ());

  outMgr::ctrl.removeOutput = true; // discard generated output
  exit (EXIT_FAILURE);
}



void
yox_internal_fail (const std::string& srcfile,
                   int lineno,
                   const std::string& msg)
{
  std::cerr << "Ox: internal failure: "
               "file " << srcfile << ", "
               "line " << lineno << ": " << msg;

  outMgr::ctrl.removeOutput = true; // discard generated output
  exit (EXIT_FAILURE);
}



void
yox_internal_fail (const std::string& srcfile,
                   int lineno,
                   std::ostringstream& msg_sstr)
{
  const std::string& msg = msg_sstr.str ();
//msg_sstr.str (std::string ());
//msg_sstr.clear (); // reset the stringstream

  std::cerr << "Ox: internal failure: "
               "file " << srcfile << ", "
               "line " << lineno << ": " << msg;

  outMgr::ctrl.removeOutput = true; // discard generated output
  exit (EXIT_FAILURE);
}



void
yox_alloc_fail (const std::string& srcfile, int lineno)
{
  std::cerr << "Ox: memory allocation failure: "
               "file " << srcfile << ", "
               "line " << lineno << ".\n";

  outMgr::ctrl.removeOutput = true; // discard generated output
  exit (EXIT_FAILURE);
}



void
agc_syntax_error (const AGC::location& loc, const std::string& msg)
{
  yox_message (loc, nullptr, msg.c_str ());
  std::cerr << "\n"; // skeleton parser error messages don't end with '\n'
  outMgr::ctrl.removeOutput = true; // discard generated output
}



void
agc_syntax_error (const AGC::location& loc, std::ostringstream& msg_sstr)
{
  std::string msg = msg_sstr.str ();
  msg_sstr.str (std::string ());
  msg_sstr.clear (); // reset the stringstream

  yox_message (loc, nullptr, msg.c_str ());
  std::cerr << "\n"; // skeleton parser error messages don't end with '\n'
  outMgr::ctrl.removeOutput = true; // discard generated output
}
