/* $Id: strip.cc 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <ios>
#include <sstream>

#if defined(WITH_REFLEX) && defined(_WIN32)
#include <reflex/input.h>
#endif

#include "agc-Lfile.hh"
#include "agc-strip-lexer.hh"
#include "fileMgr.hh"
#include "messages.hh"
#include "scan.hh"

void
stripInputFiles (int debug)
{
  AGC::lexer::value_type yylval;
  AGC::location yylloc;

  yoxControl& ctrl = outMgr::ctrl;

  ctrl.buildHeaderFile = false;

  yylloc.initialize (&Yfile.inName);

  // open the input Y-file
#if defined(WITH_REFLEX) && defined(_WIN32)
  FILE* yoxInGF = fopen (Yfile.inName.c_str (), "rb");
  if (yoxInGF == nullptr)
#else // defined(WITH_FLEX) || defined(WITH_REFLEX) && !defined(_WIN32)
  std::ifstream yoxInG (Yfile.inName);
  if (!yoxInG.is_open ())
#endif
    {
      const AGC::location loc (&Yfile.inName);
      yox_fatal (loc, "opening " + Yfile.inName + " for input failed\n");
    }

#if defined(WITH_REFLEX) && defined(_WIN32)
  reflex::BufferedInput::dos_streambuf buf (yoxInGF); // "\r\n" -> '\n'
  std::istream yoxInG (&buf);
  if (!yoxInG.good ())
    {
      const AGC::location loc (&Yfile.inName);
      yox_fatal (loc, "opening " + Yfile.inName + " for CRLF->LF failed\n");
    }
#endif

  // open the output Y-file
  yoxOutG.open (Yfile.outName);

  yoxOutG.setCurPos (yylloc.begin);

  yoxOutG.gen ().put ("%{\n"
                      "#define " + yoxPrefix ("init") + "()\n"
                      "void " + yoxPrefix ("abort") + "(void) {};\n"
                      "%}\n");

  AGC::STRIP::Lexer agc_strip_lexer (Yfile, yoxInG);
  agc_strip_lexer.set_debug (debug >> 1);

  (void)agc_strip_lexer.yylex (&yylval, &yylloc);

  // close the output Y-file
  yoxOutG.close ();

  // close the input Y-file
#if defined(WITH_REFLEX) && defined(_WIN32)
  fclose (yoxInGF);
#else // defined(WITH_FLEX) || defined(WITH_REFLEX) && !defined(_WIN32)
  yoxInG.close ();
#endif

  for (Lfile_specIO& currentLfile : Lfile)
    {
      switch (currentLfile.scanner)
        {
        case LEX_SPEC_LEXER:
          break;
        case GEN_SPEC_LEXER:
          break;
        default:
          yox_internal_fail (__FILE__, __LINE__,
                             "unexpected Lexer " +
                                 std::to_string (currentLfile.scanner) + "\n");
        }

      yylloc.initialize (&currentLfile.inName);

      // open the (next) input L-file
#if defined(WITH_REFLEX) && defined(_WIN32)
      FILE* yoxInLF = fopen (currentLfile.inName.c_str (), "rb");
      if (yoxInLF == nullptr)
#else // defined(WITH_FLEX) || defined(WITH_REFLEX) && !defined(_WIN32)
      std::ifstream yoxInL (currentLfile.inName);
      if (!yoxInL.is_open ())
#endif
        {
          const AGC::location loc (&currentLfile.inName);
          yox_fatal (loc,
                     "opening " + currentLfile.inName + " for input failed\n");
        }

#if defined(WITH_REFLEX) && defined(_WIN32)
      reflex::BufferedInput::dos_streambuf buf (yoxInLF); // "\r\n" -> '\n'
      std::istream yoxInL (&buf);
      if (!yoxInL.good ())
        {
          const AGC::location loc (&currentLfile.inName);
          yox_fatal (loc,
                     "opening " + currentLfile.inName +
                         " for CRLF->LF failed\n");
        }
#endif

      // open the (next) output L-file
      yoxOutL.open (currentLfile.outName);

      yoxOutL.setCurPos (yylloc.begin);

      AGC::STRIP::Lexer agc_strip_lexer (currentLfile, yoxInL);
      agc_strip_lexer.set_debug (debug >> 1);

      (void)agc_strip_lexer.yylex (&yylval, &yylloc);

      // close the current output L-file
      yoxOutL.close ();

      // close the current input L-file
#if defined(WITH_REFLEX) && defined(_WIN32)
      fclose (yoxInLF);
#else // defined(WITH_FLEX) || defined(WITH_REFLEX) && !defined(_WIN32)
      yoxInL.close ();
#endif
    }
}
