/* $Id: trav.cc 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2010-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <sstream>

#include <climits>
#include <list>

#include "location.hh"

#include "codegen.hh"
#include "ctc.hh"
#include "joinTabs.hh"
#include "messages.hh"
#include "opts.hh"
#include "scan.hh"
#include "symTrans.hh"
#include "trav.hh"



struct travSegElem
{
  attrbRefSeg orderRevCondRefSeg;
  attrbRefSeg actionRefSeg;

  travSegElem ()
      : orderRevCondRefSeg (),
        actionRefSeg ()
  {}
};

typedef std::list<travSegElem> travSegList;

static travSegElem travSegBuf;



struct travSpec
{
  bool RL;   /* @righttoleft */
  bool LR;   /* @lefttoright */
  bool PRE;  /* @preorder    */
  bool POST; /* @postorder   */
  bool DA;   /* @disable     */

  travSpec ()
      : RL (false),
        LR (false),
        PRE (false),
        POST (false),
        DA (false)
  {}
};

// expansion increment for the traversal space
#define MAXNTRAVSINCR 16

struct travElem
{
  std::string id;
  AGC::location loc;
  travSpec spec;
  attrbRefSeg dirRevCondRefSeg;
  int nActions;
  travSegList TSBlist;

  travElem ()
      : id (std::string ()),
        loc (AGC::location ()),
        spec (travSpec ()),
        dirRevCondRefSeg (),
        nActions (0),
        TSBlist (travSegList ())
  {}

  travElem (std::string i, AGC::location l)
      : id (i),
        loc (l),
        spec (travSpec ()),
        dirRevCondRefSeg (),
        nActions (0),
        TSBlist (travSegList ())
  {}
};

// temporary: travSp [0] is travArr [-1]
static std::vector<travElem> travSp (1 + MAXNTRAVSINCR, travElem ());

// temporary: for travArr [-1] compatibility
static travElem* travArr = &travSp[1];

static int travNumSeq = 0;

static travElem travBuf;



static int maxNactions = 0;

int currentTrav;



bool
travSpecConflict (void)
{
  return (travBuf.spec.RL && travBuf.spec.LR) ||
         (travBuf.spec.PRE && travBuf.spec.POST);
}



void
newTrav (const AGC::location& loc, const std::string& id)
{
  int travIdx = travLookup (id);

  if (travIdx != -1)
    yox_error (loc,
               "duplicate traversal name: \"" + id + "\" "
                   "(line " +
                   std::to_string (travArr[travIdx].loc.begin.line) +
                   ")"
                   ".\n");

  if (!travBuf.spec.RL && !travBuf.spec.LR)
    travBuf.spec.LR = true;
  if (!travBuf.spec.PRE && !travBuf.spec.POST)
    travBuf.spec.POST = true;

  const int size = (int)travSp.size ();
  if (travNumSeq == size - 1)
    try
      {
        travSp.resize (size + MAXNTRAVSINCR, travElem ());
        travArr = &travSp[1]; // temporary: for travArr [-1] compatibility
      }
    catch (std::bad_alloc const&)
      {
        // reallocation of traversal space failed
        yox_alloc_fail (__FILE__, __LINE__);
      }

  travArr[travNumSeq] = travElem (id, loc);
  travArr[travNumSeq].spec = travBuf.spec;
  travNumSeq++;
}



void
clearTravBuf (void)
{
  travBuf.spec.RL = false;
  travBuf.spec.LR = false;
  travBuf.spec.PRE = false;
  travBuf.spec.POST = false;
  travBuf.spec.DA = false;
}



int
travLookup (const std::string& id)
{
  for (int i = 0; i < travNumSeq; i++)
    if (id == travArr[i].id)
      return i;
  return -1;
}



void
travSetPreorder (void)
{
  travBuf.spec.PRE = true;
}



void
travSetPostorder (void)
{
  travBuf.spec.POST = true;
}



void
travSetLeftToRight (void)
{
  travBuf.spec.LR = true;
}



void
travSetRightToLeft (void)
{
  travBuf.spec.RL = true;
}



void
travSetDisable (void)
{
  travBuf.spec.DA = true;
}



bool
travDisabled (void)
{
  return travBuf.spec.DA;
}



std::string errTravID ("__undeclared__");



void
clearErrorTrav (void)
{
  travArr[-1].id = errTravID;
  travArr[-1].nActions = 0;
  travArr[-1].TSBlist.clear ();
}



void
prepareDataStructures2 (void)
{
  clearTravBuf ();
  clearErrorTrav ();
}



void
codeGen::putTravCodePerProd (void)
{
  if (travNumSeq == 0)
    return;

  gen ();
  put ("case ");
  put (prodNumSeq - 1);
  put (":  /***Yacc rule ");
  put (prodNumSeq - 1);
  put ("***/\n");

  if (travNumSeq > 1)
    put ("\tswitch(travN)\t{\n");

  for (int i = 0; i < travNumSeq; i++)
    {
      if (travNumSeq > 1)
        {
          put ("\t\tcase "); put (i); put (":\n");
        }

      put ("\t\t\tswitch(pass)\t{\n"
           "\t\t\t\tcase 0:\n"
           "righttoleft = ");

      travElem& the_trav = travArr[i];
      if (!the_trav.dirRevCondRefSeg.buffer.empty ())
        {
          if (!the_trav.spec.LR)
            put ("!");

          put ("(bool) ("); // @revdirection expr might not be of type 'bool'

          translateAndOutput (trav,
                              the_trav.dirRevCondRefSeg,
                              attrbPair (0 /*don't care*/, 0 /*don't care*/));

          gen ();
          put (")");
        }
      else
        {
          if (the_trav.spec.LR)
            put ("false");
          else
            put ("true");
        }

      put (";\n");

      the_trav.dirRevCondRefSeg.clear ();

      travSegList& TSBlist = the_trav.TSBlist;

      travSegList::const_iterator iter = TSBlist.begin ();
      for (int j = 0; iter != TSBlist.end (); iter++, j++)
        {
          if (!iter->orderRevCondRefSeg.buffer.empty ())
            {
              gen ();
              put ("if (");

              if (!the_trav.spec.POST)
                put ("!");

              put ("(bool) ("); // @revorder expr might not be of type 'bool'

              translateAndOutput (trav,
                                  iter->orderRevCondRefSeg,
                                  attrbPair (0 /*don't care*/,
                                             0 /*don't care*/));

              gen ();
              put (")");

              put (") " + yoxPrefix ("SetCond") + "("); put (j); put (")\n");
            }
          else
            {
              if (!the_trav.spec.POST)
                {
                  put (yoxPrefix ("SetCond") + "("); put (j); put (")\n");
                }
            }
        } // for j

      if (!TSBlist.empty ())
        {
          put ("#if (__APPLE__ && __clang__ && defined "
                   "(__apple_build_version__) && "
                   "__clang_major__ > 11) || \\\n"
               "    (__linux__ && __GNUC__ > 6)\n"
               "/* Apple Xcode 10 & 11 gcc (tested):\n"
               "   warning: declaration does not declare anything\n"
               "   GNU 4, 5 & 6 gcc (tested):\n"
               "   error: expected primary-expression before "
                   "\'__attribute__\'\n"
               " */\n"
               "__attribute__((fallthrough));\n"
               "#elif defined (S_SPLINT_S)\n"
               "/*@fallthrough@*/\n"
               "#elif defined (lint)\n"
               "NOTE(FALLTHROUGH)\n"
               "#endif\n"
               "\t\t\t\tcase 1:\n");

          travSegList::const_iterator iter = TSBlist.begin ();
          for (int j = 0; iter != TSBlist.end (); iter++, j++)
            {
              put ("if (" + yoxPrefix ("Cond") + "(");
              put (j);
              put (") != pass) {");;

              translateAndOutput (trav,
                                  iter->actionRefSeg,
                                  attrbPair (0 /*don't care*/,
                                             0 /*don't care*/));

              gen ();
              put ("}\n");
            } // for j

          put ("\t\t\t\tbreak;\n");
        }

      gen (); // TES; not needed?
      put ("\t\t\t\t\t}\n");

      if (travNumSeq > 1)
        put ("\t\tbreak;\n");

      if (maxNactions < the_trav.nActions)
        maxNactions = the_trav.nActions;
      the_trav.nActions = 0;
      the_trav.TSBlist.clear ();
    } // for i

  if (travNumSeq > 1)
    put ("\t\t\t}\n");

  put ("\nbreak;\n");

  clearErrorTrav ();
}



void
codeGen::putTravDefines (void)
{
  gen ();
  put ("\n\n\n");

  if (maxNactions <= CHAR_BIT * (int)sizeof (long long))
    {
      if (maxNactions <= CHAR_BIT * (int)sizeof (char))
        put ("#define " + yoxPrefix ("CondStg") + " unsigned char conds;\n");
      else if (maxNactions <= CHAR_BIT * (int)sizeof (short))
        put ("#define " + yoxPrefix ("CondStg") + " unsigned short conds;\n");
      else if (maxNactions <= CHAR_BIT * (int)sizeof (int))
        put ("#define " + yoxPrefix ("CondStg") + " unsigned int conds;\n");
      else if (maxNactions <= CHAR_BIT * (int)sizeof (long))
        put ("#define " + yoxPrefix ("CondStg") + " unsigned long conds;\n");
      else if (maxNactions <= CHAR_BIT * (int)sizeof (long long))
        put ("#define " + yoxPrefix ("CondStg") +
             " unsigned long long conds;\n");

      put ("#define " + yoxPrefix ("ClearConds") +
           " {travStack[travStkPtr].conds = 0;}\n");
      put ("#define " + yoxPrefix ("SetCond") +
           "(n) {travStack[travStkPtr].conds |= (1<<(n));}\n");
      put ("#define " + yoxPrefix ("Cond") +
           "(n) (((travStack[travStkPtr].conds & (1<<(n)))!=0)?1:0)\n");
    }
  else
    {
      put ("#define " + yoxPrefix ("NCW") + " ceil(");
      put (maxNactions); put (",sizeof(unsigned int));\n");
      put ("#define " + yoxPrefix ("CondStg") + " unsigned int conds[" +
           yoxPrefix ("NCW") + "];\n");
      put ("#define " + yoxPrefix ("ClearConds") + " {for (int i=0;i<" +
           yoxPrefix ("NCW") + ";i++)"
           "travStack[travStkPtr].conds[i]=0;}\n");
      put ("#define " + yoxPrefix ("SetCond") +
           "(n) {travStack[travStkPtr].conds[(n)/sizeof(unsigned int)] "
           "|= (1<<(n%sizeof(unsigned int)));}\n");
      put ("#define " + yoxPrefix ("Cond") +
           "(n) ((travStack[travStkPtr].conds[(n)/sizeof(unsigned int)] "
           "& (1<<(n%sizeof(unsigned int))))?1:0)\n");
    }
}



void
wrapUpTravDirCond (void)
{
  travArr[currentTrav].dirRevCondRefSeg = rsBuffer.get ();
}



void
wrapUpTravOrderCond (void)
{
  travSegBuf.orderRevCondRefSeg = rsBuffer.get ();
}



static void
clearTSBbuf (void)
{
  travSegBuf.orderRevCondRefSeg.clear ();
  travSegBuf.actionRefSeg.clear ();
}



void
wrapUpTravAction (void)
{
  travSegBuf.actionRefSeg = rsBuffer.get ();

  try
    {
      travArr[currentTrav].nActions++;
      travArr[currentTrav].TSBlist.push_back (travSegBuf);
    }
  catch (...)
    {
      // expansion of the traversal segment buffer list failed
      yox_alloc_fail (__FILE__, __LINE__);
    }

  clearTSBbuf ();
}



static const char* const skeletonDoTraversalHeader[] = {
  "\n",
  "\n",
  "\n",
  "%@struct %oTravStackItem {%oGNT *node;\n",
  "                         bool isReady;\n",
  "%@                         %oCondStg\n",
  "                        };\n",
  "\n",
  "\n",
  "\n",
  "%@static void %oDoTraversals(%oGNT *rootNode)\n",
  "%@{struct %oTravStackItem *travStack;\n",
  " long travStkPtr,afterTravStack;\n",
  "%@ %oGNT *travNode,**cLptr2;\n",
  " bool righttoleft=false; int pass;\n",
  " int i;\n",
  "\n",
  "%@ if (!%oYok) return;\n",
  " if ((travStack =\n",
  "%@                 ((struct %oTravStackItem *)\n",
  "%@                  calloc((size_t)%oTravStackMaxSize,\n",
  "%@                         (size_t)sizeof(struct %oTravStackItem)\n",
  "                        )\n",
  "                 )\n",
  "     )\n",
  "     ==\n",
  "%@     (struct %oTravStackItem *) NULL\n",
  "    )\n",
  "%@    %ofatal(\"malloc error in Ox Traversal Stack allocation\\n\");\n",
  "\n",
  "%@afterTravStack = %oTravStackMaxSize;\n",
  "%@#define %oTSBottom 1L\n",
  "\n",
  "\n",
}; // skeletonDoTraversalHeader[]

static const char* const skeletonTravStackMgr[] = {
  "travStkPtr = 1L;\n",
  "travStack[travStkPtr].node = rootNode;\n",
  "travStack[travStkPtr].isReady = false;\n",
  "%@%oClearConds\n",
  "\n",
  "%@while(travStkPtr >= %oTSBottom)\n",
  "  {travNode = travStack[travStkPtr].node;\n",
  "   if (travStack[travStkPtr].isReady)\n",
  "      {pass = 1;\n",
  "       goto travSwitch;\n",
  "popTravStack:\n",
  "       travStkPtr--;\n",
  "      }\n",
  "   else\n",
  "      {pass = 0;\n",
  "       goto travSwitch;\n",
  "pushTravStack:\n",
  "       travStack[travStkPtr].isReady = true;\n",
  "       if (travNode->prodNum!=0)\n",
  "          {if (righttoleft)\n",
  "             {cLptr2 = travNode->cL;\n",
  "              i = travNode->cLlen;\n",
  "              while ((i--)>0)\n",
  "                {if (++travStkPtr == afterTravStack)\n",
  "                    {/* realloc `travStack' data structure */\n",
  "%@                     %oTravStackMaxSize += 500L;\n",
  "%@                     if ((travStack = (struct %oTravStackItem *)\n",
  "                          realloc(travStack,\n",
  "%@                                  "
      "(size_t)((size_t)%oTravStackMaxSize*sizeof(struct %oTravStackItem)))\n",
  "%@                         ) == (struct %oTravStackItem *) NULL)\n",
  "%@                        %ofatal(\"malloc error in Ox Traversal Stack "
      "allocation\\n\");\n",
  "%@                     afterTravStack = %oTravStackMaxSize;\n",
  "                    }\n",
  "                 travStack[travStkPtr].node = *cLptr2;\n",
  "                 travStack[travStkPtr].isReady = false;\n",
  "%@                 %oClearConds\n",
  "                 cLptr2++;\n",
  "                }\n",
  "             } /* right to left */\n",
  "          else  /* left to right */\n",
  "             {i = travNode->cLlen;\n",
  "              cLptr2 = travNode->cL + i;\n",
  "              while ((i--)>0)\n",
  "                {cLptr2--;\n",
  "                 if (++travStkPtr == afterTravStack)\n",
  "                    {/* realloc `travStack' data structure */\n",
  "%@                     %oTravStackMaxSize += 500L;\n",
  "%@                     if ((travStack = (struct %oTravStackItem *)\n",
  "                          realloc(travStack,\n",
  "%@                                  "
      "(size_t)((size_t)%oTravStackMaxSize*sizeof(struct %oTravStackItem)))\n",
  "%@                         ) == (struct %oTravStackItem *) NULL)\n",
  "%@                        %ofatal(\"malloc error in Ox Traversal Stack "
      "allocation\\n\");\n",
  "%@                     afterTravStack = %oTravStackMaxSize;\n",
  "                    }\n",
  "                  travStack[travStkPtr].node = *cLptr2;\n",
  "                  travStack[travStkPtr].isReady = false;\n",
  "%@                  %oClearConds\n",
  "                }\n",
  "             } /* left to right */\n",
  "          }\n",
  "      } /* else */\n",
  "   continue;\n",
  "travSwitch:\n",
  "\t\t\t\tswitch(travNode->prodNum)\t{\n",
}; // skeletonTravStackMgr[]

static const char* const skeletonDoTraversalTrailer[] = {
  "\t\t\t\t\t\t\t\t} /* switch */\n",
  "   if (pass!=0) goto popTravStack; else goto pushTravStack;\n",
  "  } /* while */\n",
  " } /* for */\n",
  "%@} /* %oDoTraversals */\n",
  "\n",
}; // skeletonDoTraversalTrailer[]

void
codeGen::putTravCode ()
{
  if (travNumSeq == 0)
    return;

  putTravDefines ();

  putStringArray (skeletonDoTraversalHeader,
                  sizeof (skeletonDoTraversalHeader));

  gen ();
  put ("for (int travN=0; travN<"); put (travNumSeq); put ("; travN++) {\n");

  for (int i = 0; i < travNumSeq; i++)
    if (travArr[i].spec.DA)
      {
        put ("if (travN == "); put (i); put (") continue;\n");
      }

  putStringArray (skeletonTravStackMgr, sizeof (skeletonTravStackMgr));

  append (tempTravOut);

  putStringArray (skeletonDoTraversalTrailer,
                  sizeof (skeletonDoTraversalTrailer));
}



codeGen&
codeGen::putDoTraversals (const std::string& rootNodeReference)
{
  if (travNumSeq > 0)
    {
      gen ();
      put ("\t\t    " + yoxPrefix ("DoTraversals") + "(");
      put (rootNodeReference);
      put (");\n");
    }

  return *this;
}
