/* $Id: treeviz.cc 2841 2022-12-13 01:08:03Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2019-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include "codegen.hh"
#include "ctc.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "symTrans.hh"
#include "treeviz.hh"



// NIO == Node-Image Object

// NIO pseudo attribute # (symbol table pseudo index):
const int NIOsymAttrbNum = -2;


// NIO pseudo attribute name:
#define NIO_attribute_name "NIO"
// Q: allow this to be defined by the evaluator author?


std::string NIOsym; // attrbSymTable 'NIO' pseudo attribute



// install & memoize the `NIO' attribute
void
setupNIOsym (void)
{
  NIOsym = stringLookup (NIO_attribute_name, true);

  const AGC::location loc; // dummy location

  attrbAutoDef (loc, NIOsym, undefined, false);
}



void
codeGen::putGenTreeVizLeafCall (const std::string& token)
{
  put ("\n" + yoxPrefix ("GenTreeVizLeaf") + "(");

  if (token == omittedRetSym)
    put ("yytext,&");
  else
    put ("\"" + token + "\",&");

  // "yylval(.|->)yyyOxAttbs.semEl->node->"
  putImmediateAttrbRef (attrbPair (0, NIOsymAttrbNum), false);

  put (");");
}



void
codeGen::putGenTreeVizNodeCall (const std::string& lhs_symbol)
{
  put ("\n" + yoxPrefix ("GenTreeVizNode") + "(" +
       std::to_string (prodBufferPointer2 - 1) + ","
       "\"" + lhs_symbol + "\","
       "&$<" + yoxPrefix ("OxAttrbs") + ">$.semEl->node->" + NIOsym);
  putRHSsemStkElRefList (".semEl->node->" + NIOsym);
  put (");");
}



void
codeGen::putTreeVizId (void)
{
  put ("#define YYYOX_TREE 1\n");
}



void
codeGen::putTreeVizExtDecl (void)
{
  put ("extern void " + yoxPrefix ("init") + " (void);\n" +
       "extern Agraph_t *" + yoxPrefix ("TreeVizGraph") + ";\n");
}



void
codeGen::putGenTreeVizLeafExtDecl (void)
{
  put ("extern void " + yoxPrefix ("GenTreeVizLeaf") +
       " (const char *, Agnode_t **);\n");
}



void
codeGen::putTreeVizGraphvizIncl (void)
{
  put ("\n\n\n#include <graphviz/types.h>\n\n\n");
}



void
codeGen::putGenTreeVizNodeFwdDecl (void)
{
  put ("\n"
       "static void " + yoxPrefix ("GenTreeVizNode") +
       " (int, const char *, Agnode_t **,...);\n");
}



void
codeGen::putTreeVizUserDefPrefix (void)
{
  put ("#define yyyTreeVizGraph " + yoxPrefix ("TreeVizGraph") + "\n");
}



#ifndef AGRAPH_RANKDIR
#define AGRAPH_RANKDIR 2
// 1 = "TB" derivation tree root at top
// 2 = "LR" 90 degree rotation counter-clockwise from "TB"
// 3 = "BT" derivation tree root at bottom
// 4 = "RL" 90 degree rotation clockwise from "TB"
#endif

static const char* const pttC[] = {
  "\n",
  "%@Agraph_t *%oTreeVizGraph = (Agraph_t *) NULL;\n",
  "\n",
  "%@static Agsym_t *%oAgsym_node_label;\n",
  "%@static Agsym_t *%oAgsym_node_shape;\n",
  "\n",
  "\n",
  "\n",
  "/* LR bottom-up derivation tree edge comparison, for ordered traversal. "
      "*/\n",
  "#if defined (__cplusplus)\n",
  "#include <cassert>\n",
  "#else\n",
  "#include <assert.h>\n",
  "#endif\n",
  "%@static int %oAgedgeCmp (__attribute__((unused)) Dict_t * d,\n",
  "                         Agedge_t *e0, Agedge_t *e1,\n",
  "                         __attribute__((unused)) Dtdisc_t * disc\n",
  "                        )\n",
  "  {\n",
  "#ifndef __GNUC__\n",
  "   (void) d; (void) disc;\n",
  "#endif\n",
  "   assert (e0 && e1);\n",
  "   if (AGSEQ(e0) < AGSEQ(e1)) return -1;\n",
  "   if (AGSEQ(e0) > AGSEQ(e1)) return  1;\n",
  "   return  0;\n",
  "  }\n",
  "\n",
  "\n",
  "\n",
  "/* `Ag_mainedge_seq_disc' is not in the public graphviz headers */\n",
  "extern Dtdisc_t Ag_mainedge_seq_disc;\n",
  "\n",
  "%@void %oinit (void)\n",
  "  {static const char graphLabel[] = \"Ox-generated evaluator parse "
      "tree\";\n",
  "   static const char graphRankdir[] =",
#if AGRAPH_RANKDIR == 1
  " \"TB\"",
#elif AGRAPH_RANKDIR == 2
  " \"LR\"",
#elif AGRAPH_RANKDIR == 3
  " \"BT\"",
#elif AGRAPH_RANKDIR == 4
  " \"RL\"",
#else
#warning "undefined choice for `graphRankdir', using \"LR\"."
#undef AGRAPH_RANKDIR
#define AGRAPH_RANKDIR 2
  " \"LR\"",
#endif
  ";\n",
  "\n",
  "   static bool initDone = false;\n",
  "   if (initDone) return; initDone = true;\n",
  "\n",
  "   /* replace the default edge compare function, which */\n",
  "   /* orders out-edges by the timestamp of the head node */\n",
  "%@   Ag_mainedge_seq_disc.comparf = (Dtcompar_f)%oAgedgeCmp;\n",
  "\n",
  "%@   %oTreeVizGraph = agopen ((char *) graphLabel, Agdirected, NULL);\n",
  "\n",
  "   /* default settings for the graph */\n",
  "%@   agattr (%oTreeVizGraph, AGRAPH, (char *) \"ordering\",    (char *) "
      "\"out\");\n",
  "%@   agattr (%oTreeVizGraph, AGRAPH, (char *) \"outputorder\", (char *) "
      "\"breadthfirst\");\n",
  "%@   agattr (%oTreeVizGraph, AGRAPH, (char *) \"label\",       (char *) "
      "graphLabel);\n",
  "%@   agattr (%oTreeVizGraph, AGRAPH, (char *) \"labeljust\",   (char *) "
      "\"left\");\n",
  "%@   agattr (%oTreeVizGraph, AGRAPH, (char *) \"rankdir\",     (char *) "
      "graphRankdir);\n",
  "%@   agattr (%oTreeVizGraph, AGRAPH, (char *) \"splines\",     (char *) "
      "\"polyline\");\n",
  "\n",
  "   /* default settings for nodes */\n",
  "%@   %oAgsym_node_label = agattr (%oTreeVizGraph, AGNODE, (char *) "
      "\"label\", (char *) \"\"); /* unspecified */\n",
  "%@   %oAgsym_node_shape = agattr (%oTreeVizGraph, AGNODE, (char *) "
      "\"shape\", (char *) \"\"); /* unspecified */\n",
  "%@   agattr (%oTreeVizGraph, AGNODE, (char *) \"style\", (char *) "
      "\"bold\");\n",
  "\n",
  "   /* default settings for edges */\n",
  "%@   agattr (%oTreeVizGraph, AGEDGE, (char *) \"arrowhead\", (char *) "
      "\"none\"); /* no arrowhead */\n",
  "%@   agattr (%oTreeVizGraph, AGEDGE, (char *) \"style\",     (char *) "
      "\"dashed,bold\");\n",
  "  }\n",
  "\n",
  "\n",
  "\n",
  "%@static void %oGVlabelTrans (char *trans, const char *cp)\n",
  "  {while (*cp != '\\0')\n",
  "     {int i = (int) strcspn (cp, \"{}<>| \\\\\\\"\");\n",
  "      if (i > 0) {(void) strncpy (trans, cp, i); cp+=i; trans+=i;}\n",
  "      if (*cp == '\\0') break;\n",
  "      *trans++ = '\\\\'; *trans++ = *cp++;\n",
  "     }\n",
  "   *trans = '\\0';\n",
  "  }\n",
  "\n",
  "\n",
  "\n",
  "/* NIO == Node-image Object */\n",
  "\n",
  "%@void %oGenTreeVizLeaf (const char *label, Agnode_t **lval_NIO)\n",
  "  {Agnode_t *leaf_node;\n",
  "   char *trans;\n",
  "\n",
  "   static int leaf_node_num = 0;\n",
  "   static char leaf_node_name[20];\n",
  "   (void) snprintf (leaf_node_name, sizeof (leaf_node_name), \"Leaf%d\", "
      "++leaf_node_num);\n",
  "\n",
  "%@   if (%oTreeVizGraph == (Agraph_t *) NULL) %oinit();\n",
  "\n",
  "%@   leaf_node = agnode (%oTreeVizGraph, leaf_node_name, true);\n",
  "#if defined (__cplusplus)\n",
  "   try {\n",
  "      trans = new char [strlen (label)*2]; // worst case translation "
      "length\n",
  "   } catch (std::bad_alloc) {\n",
  "%@      %ofatal (\"error in %oGenTreeVizLeaf() space allocation\\n\");\n",
  "   }\n",
  "#else\n",
  "   trans = (char *) malloc (strlen (label)*2); /* worst case translation "
      "length */\n",
  "%@   if (trans == (char *) NULL) %ofatal (\"malloc error in "
      "%oGenTreeVizLeaf()\\n\");\n",
  "#endif\n",
  "%@   %oGVlabelTrans (trans, label);\n",
  "%@   agxset (leaf_node, %oAgsym_node_label, trans);\n",
  "#if defined (__cplusplus)\n",
  "   delete[] trans;\n",
  "#else\n",
  "   free (trans);\n",
  "#endif\n",
  "%@   agxset (leaf_node, %oAgsym_node_shape, (char *) \"Mrecord\");\n",
  "\n",
  "   *lval_NIO = leaf_node;\n",
  "  }\n",
  "\n",
  "\n",
  "\n",
  "%@static void %oGenTreeVizNode (int RHSlength, const char *label, Agnode_t "
      "**val_NIO, ...)\n",
  "  {Agnode_t *LHSnode;\n",
  "   char *trans;\n",
  "\n",
  "   static int node_num = 0;\n",
  "   static char node_name[100];\n",
  "   (void) snprintf (node_name, sizeof (node_name), \"Node%d\", "
      "++node_num);\n",
  "\n",
  "%@   if (%oTreeVizGraph == (Agraph_t *) NULL) %oinit();\n",
  "\n",
  "%@   LHSnode = agnode (%oTreeVizGraph, node_name, true);\n",
  "#if defined (__cplusplus)\n",
  "   try {\n",
  "      trans = new char [strlen (label)*2]; // worst case translation "
      "length\n",
  "   } catch (std::bad_alloc) {\n",
  "%@      %ofatal (\"error in %oGenTreeVizNode() space allocation\\n\");\n",
  "   }\n",
  "#else\n",
  "   trans = (char *) malloc (strlen (label)*2); /* worst case translation "
      "length */\n",
  "%@   if (trans == (char *) NULL) %ofatal (\"malloc error in "
      "%oGenTreeVizNode()\\n\");\n",
  "#endif\n",
  "%@   %oGVlabelTrans (trans, label);\n",
  "%@   agxset (LHSnode, %oAgsym_node_label, trans);\n",
  "#if defined (__cplusplus)\n",
  "   delete[] trans;\n",
  "#else\n",
  "   free (trans);\n",
  "#endif\n",
  "%@   agxset (LHSnode, %oAgsym_node_shape, (char *) \"record\");\n",
  "\n",
  "   if (RHSlength > 0)\n",
  "      {va_list ap;\n",
  "\n",
#if AGRAPH_RANKDIR == 2 // graphRankdir == "LR" 
  "#if (__STDC_VERSION__ >= 199901L) && !defined (__cplusplus) /* ISO C99 (or "
      "later) C */\n",
  "       Agnode_t *RHS[RHSlength];\n",
  "#else /* C++ and C prior to C99 */\n",
  "       Agnode_t **RHS = (Agnode_t **) NULL;\n",
  "#if defined (__cplusplus)\n",
  "       try {\n",
  "          RHS = new Agnode_t * [RHSlength];\n",
  "       } catch (std::bad_alloc) {\n",
  "%@          %ofatal (\"error in %oGenTreeVizNode() space "
      "allocation\\n\");\n",
  "       }\n",
  "#else\n",
  "       RHS = (Agnode_t **)calloc (sizeof (Agnode_t *), RHSlength);\n",
  "       if (RHS == (Agnode_t **) NULL)\n",
  "%@          %ofatal (\"calloc error in %oGenTreeVizNode()\\n\");\n",
  "#endif\n",
  "#endif\n",
  "\n",
  "       va_start (ap, val_NIO);\n",
  "       for (int i = 0; i < RHSlength; i++)\n",
  "         {Agnode_t **RHS_NIO = va_arg (ap, Agnode_t **);\n",
  "\n",
  "          if (RHS_NIO == (Agnode_t **) NULL)\n",
  "             /* the i-th RHS (child) is the `error' token */\n",
  "%@             %oGenTreeVizLeaf (\"error\",&RHS[i]);\n",
  "          else\n",
  "             RHS[i] = *RHS_NIO; /* the i-th RHS (child) node */\n",
  "         }\n",
  "       va_end (ap);\n",
  "\n",
  "       for (int i = RHSlength -1; i >= 0; i--)\n",
  "         {/* there is an unfortunate interaction between the 'rankdir=LR' "
      "and\n",
  "             'ordering=out' attributes in the 'dot' layout algorithm that "
      "orders\n",
  "             child nodes reversed from the 'rotate the top-down derivation "
      "tree\n",
  "             90 degrees counter-clockwise' expected visual order, so add "
      "the RHS\n",
  "             nodes in reverse order */\n",
  "\n",
  "          static int edge_num = 0;\n",
  "          static char edge_name[20];\n",
  "          (void) snprintf (edge_name, sizeof (edge_name), \"Edge%d\", "
      "++edge_num);\n",
  "\n",
  "%@          agedge (%oTreeVizGraph,\n",
  "                  LHSnode, /* the LHS (parent) node */\n",
  "                  RHS[i],\n",
  "                  edge_name, true\n",
  "                 );\n",
  "         }\n",
  "#if (__STDC_VERSION__ >= 199901L) && !defined (__cplusplus) /* ISO C99 (or "
      "later) C */\n",
  "#else /* C++ and C prior to C99 */\n",
  "#if defined (__cplusplus)\n",
  "       delete[] RHS;\n",
  "#else\n",
  "       free (RHS);\n",
  "#endif\n",
  "#endif\n",
#else // graphRankdir == "TB" or "RL" or "BT"
  "       va_start (ap, val_NIO);\n",
  "       for (int i=0; i<RHSlength; i++)\n",
  "         {Agnode_t **RHS_NIO = va_arg (ap, Agnode_t **);\n",
  "          Agnode_t *RHSnode;\n",
  "\n",
  "          static int edge_num = 0;\n",
  "          static char edge_name[20];\n",
  "          (void) snprintf (edge_name, sizeof (edge_name), \"Edge%d\", "
      "++edge_num);\n",
  "\n",
  "          if (RHS_NIO == (Agnode_t **) NULL)\n",
  "             /* the i-th RHS (child) is the `error' token */\n",
  "%@             %oGenTreeVizLeaf (\"error\", RHSnode);\n",
  "          else\n",
  "             RHSnode = *RHS_NIO; /* the i-th RHS (child) node */\n",
  "\n",
  "%@          agedge (%oTreeVizGraph,\n",
  "                  LHSnode, /* the LHS (parent) node */\n",
  "                  RHSnode, /* the i-th RHS (child) node */\n",
  "                  edge_name, true\n",
  "                 );\n",
  "         }\n",
  "       va_end (ap);\n",
#endif /* graphRankdir choices */
  "      }\n",
//"   else /* RHSlength == 0 */\n",
//"     {\n",
//"      /* any utility in having an <empty> node in the parse tree */\n",
//"      /* visualization? If so, the code to build it goes here    */\n",
//"     }\n",
  "   *val_NIO = LHSnode;\n",
  "  }\n",
  "\n",
  "\n",
  "\n",
};



void
codeGen::putTreeVizCode (void)
{
  putStringArray (pttC, sizeof (pttC));
}
