#if defined (__cplusplus)
#include <cstdlib>
#include <cstring>
#include <cstdio>
#include <climits>
#else
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>
#endif

#include "parser.h"
#include "util.h"

void
yyerror(YYLTYPE *yyloc, const char *msg)
{
    fprintf(stderr,"error @[%d.%d:%d.%d] %s\n",
                   yyloc->first_line,yyloc->first_column,
                   yyloc->last_line, yyloc->last_column,
                   msg);
}

void
eat_comment(const char *end,int *yycolno,int *yylineno) {
    switch (end[0]) {
    case '}':
        { char c;
          while((c = yyyinput()) != '}') ;
	  return;
        }
    case '*':
        { char c, lastc=' ';
          while((c = yyyinput())) {
	      if(lastc=='*' && c==')')
	          break;
              else
	          lastc = c;
          }
        }
    }
}

#define MAXSYMBOLS 1000

/* Ox-attribute maintained symbol table */
static struct symbolTableItem globalSymbolTable[MAXSYMBOLS];
int nextSym = 0;

symTable g_SymbolTable() { return (symTable)&globalSymbolTable; };

static int
addSymbol(symTable symTab, YYLTYPE *yyloc_sym, char *sym) {
    char *s;

    if (nextSym == MAXSYMBOLS) {
        yyerror(yyloc_sym, "addSymbol: MAXSYMBOLS exceeded");
        exit(-1);
    }

    if ((s = strdup(sym)) == NULL) {
        yyerror(yyloc_sym, "addSymbol: malloc() failure");
        exit(-1);
    }

    symTab[nextSym].sym = s;
    symTab[nextSym].isConstVal = false;
    symTab[nextSym].constVal   = 0;
    symTab[nextSym].isEnumLit  = false;
    symTab[nextSym].whichEnum  = NULL;
    symTab[nextSym].isTypeName = false;
    symTab[nextSym].whichType  = NULL;

    return nextSym++;
}

char *
addSymbolIfNew(symTable symTab, YYLTYPE *yyloc_sym, char *sym) {
    char *s;

    for(int i=0; i < nextSym; i++)
        if (strcmp(sym, symTab[i].sym) == 0) return symTab[i].sym;

    if (nextSym == MAXSYMBOLS) {
        yyerror(yyloc_sym, "addSymbolIfNew: MAXSYMBOLS exceeded");
        exit(-1);
    }

    if ((s = strdup(sym)) == NULL) {
        yyerror(yyloc_sym, "addSymbolIfNew: malloc() failure");
        exit(-1);
    }

    symTab[nextSym].sym = s;
    symTab[nextSym].isConstVal = false;
    symTab[nextSym].constVal   = 0;
    symTab[nextSym].isEnumLit  = false;
    symTab[nextSym].whichEnum  = NULL;
    symTab[nextSym].isTypeName = false;
    symTab[nextSym].whichType  = NULL;
    nextSym++;

    return s;
}

void
dumpSymbolTable(void) {
    static struct symbolTableItem *symTab = (symTable)&globalSymbolTable;
    for(int i=0; i < nextSym; i++) {
        fprintf(stdout, "symbol %s:", symTab[i].sym);
        if (symTab[i].isConstVal)
            fprintf(stdout, " constVal = %d", symTab[i].constVal);
        if (symTab[i].isEnumLit)
            fprintf(stdout, " isEnumLiteral = %d", symTab[i].whichEnum->constVal);
        if (symTab[i].isTypeName) {
            fputs(" isType", stdout);
            if (symTab[i].whichType != NULL)
                switch (symTab[i].whichType->tag) {
                case enumT:
                    fputs(" enum", stdout);
                    break;
                case rangeT:
                    fprintf(stdout, " range[%d .. %d]",
                                    symTab[i].whichType->type.rangeType.start,
                                    symTab[i].whichType->type.rangeType.end);
                    break;
                }
        }
        fputc('\n', stdout);
    }
}

symbol
newConst(symTable symTab, YYLTYPE yyloc_sym, char *sym, int iVal) {
    int which;

    for(which=0; which < nextSym; which++) {
        if (strcmp(sym, symTab[which].sym) == 0) {
	    if (symTab[which].isEnumLit) continue;
	    else if (symTab[which].isConstVal) {
		const char msg1[] = "newConst: [";
		const char msg2[] = "] is already a constant";
		char *msg = (char *)malloc(strlen(msg1)+strlen(sym)+strlen(msg2)+1);
		if (msg == NULL) {
		    yyerror(&yyloc_sym, "newConst: malloc() failure");
		    exit(-1);
		} else
		    yyerror(&yyloc_sym, strcat(strcat(strcpy(msg,msg1),sym),msg2));
		break;
	    } else goto done;
	}
    }
    which = addSymbol(symTab,&yyloc_sym,sym);
done:
    symTab[which].isConstVal = true;
    symTab[which].constVal   = iVal;

    return &symTab[which];
}


YYLTYPE
locSpan(YYLTYPE left, YYLTYPE right) {
    YYLTYPE loc;
    loc.first_line   = left.first_line;
    loc.first_column = left.first_column;
    loc.last_line    = right.last_line;
    loc.last_column  = right.last_column;
    return loc;
}

int
valOf(symTable symTab, YYLTYPE yyloc_sym, char *sym) {
    int which;

    for(which=0; which < nextSym; which++) {
        if (strcmp(sym, symTab[which].sym) == 0)
	    if (symTab[which].isConstVal) return symTab[which].constVal;
    }

    {
        const char msg1[] = "valOf: [";
        const char msg2[] = "] is not a constant";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym)+strlen(msg2)+1);
        if (msg == NULL) {
            yyerror(&yyloc_sym, "valOf: malloc() failure");
            exit(-1);
        } else {
            yyerror(&yyloc_sym, strcat(strcat(strcpy(msg,msg1),sym),msg2));
            return INT_MIN;
        }
    }
}

symbol
newType(symTable symTab, YYLTYPE yyloc, char *sym, typeSpecT spec) {
    int which;

    for(which=0; which < nextSym; which++)
        if (strcmp(sym, symTab[which].sym) == 0) break;

    if (symTab[which].isTypeName) {
        const char msg1[] = "newType: [";
        const char msg2[] = "] is already a typename";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym)+strlen(msg2)+1);
        if (msg == NULL) {
            yyerror(&yyloc, "newType: malloc() failure");
            exit(-1);
        } else
            yyerror(&yyloc, strcat(strcat(strcpy(msg,msg1),sym),msg2));
    }

    symTab[which].isTypeName = true;
    symTab[which].whichType  = spec;

    return &symTab[which];
}

typeSpecT
makeEnumType(YYLTYPE yyloc_list, enumListT list) {
    typeSpecT newTypeSpec = (typeSpecT)malloc(sizeof(struct typeSpec));
    if (newTypeSpec == NULL) {
        yyerror(&yyloc_list, "makeEnumType: malloc() failure");
        exit(-1);
    }

    newTypeSpec->tag = enumT;
    newTypeSpec->type.enumType = *list;

    return newTypeSpec;
}

typeSpecT
makeSubrangeType(YYLTYPE yyloc, int start, int end) {
    typeSpecT newTypeSpec = (typeSpecT)malloc(sizeof(struct typeSpec));
    if (newTypeSpec == NULL) {
        yyerror(&yyloc, "makeSubrangeType: malloc() failure");
        exit(-1);
    }

    newTypeSpec->tag = rangeT;
    newTypeSpec->type.rangeType.start = start;
    newTypeSpec->type.rangeType.end   = end;

    return newTypeSpec;
}

typeSpecT
typeOf(symTable symTab, YYLTYPE yyloc_sym, char *sym) {
    int which;

    for(which=0; which < nextSym; which++)
        if (strcmp(sym, symTab[which].sym) == 0) break;

    if (symTab[which].isTypeName)
        return symTab[which].whichType;
    else {
        const char msg1[] = "typeOf: [";
        const char msg2[] = "] is not a typename";
        char *msg = (char *)malloc(strlen(msg1)+strlen(sym)+strlen(msg2)+1);
        if (msg == NULL) {
            yyerror(&yyloc_sym, "typeOf: malloc() failure");
            exit(-1);
        }
        yyerror(&yyloc_sym, strcat(strcat(strcpy(msg,msg1),sym),msg2));
        return NULL;
    }
}

enumListT
newEnum(symTable symTab, YYLTYPE yyloc_sym, char *sym) {
    int which;

    for(which=0; which < nextSym; which++) {
        if (strcmp(sym, symTab[which].sym) == 0) {
	    if (symTab[which].isConstVal) continue;
	    else if (symTab[which].isEnumLit) {
		const char msg1[] = "newEnum: [";
		const char msg2[] = "] is already an enum literal";
		char *msg = (char *)malloc(strlen(msg1)+strlen(sym)+strlen(msg2)+1);
		if (msg == NULL) {
		    yyerror(&yyloc_sym, "newEnum: malloc() failure");
		    exit(-1);
		} else
		    yyerror(&yyloc_sym, strcat(strcat(strcpy(msg,msg1),sym),msg2));
		break;
	    } else goto done;
	}
    }
    which = addSymbol(symTab,&yyloc_sym,sym);
done:
    {
	enumListT newL = (enumListT)malloc(sizeof(struct enumList));
	if (newL == NULL) {
	    yyerror(&yyloc_sym, "newEnum: malloc() failure");
	    exit(-1);
	}

	newL->enumLit  = &symTab[which]; 
	newL->constVal = 1;
	newL->next     = NULL;

	symTab[which].isEnumLit = true;
	symTab[which].whichEnum = newL;

	return newL;
    }
}

enumListT
addToEnum(symTable symTab, YYLTYPE yyloc_sym, enumListT eL, char *sym) {
    enumListT tL = eL;
    enumListT newL = newEnum(symTab,yyloc_sym,sym);

    while (tL->next != NULL) tL = tL->next;

    newL->constVal = tL->constVal+1;

    tL->next = newL;

    return eL;
}
