/* Infix notation calculator--calc */
%code requires /* -> header file, preceding the definition of YYSTYPE */
{
#if defined (__cplusplus)
#include <cstdbool>
#else
#include <stdbool.h>
#endif

#ifndef YY_TYPEDEF_YY_SCANNER_T
# define YY_TYPEDEF_YY_SCANNER_T
typedef void* yyscan_t;
#endif

  typedef struct
  {
    bool verbose; /* print the intermediate results */
    int value;    /* value of the last computation */
    int nerrs;    /* count of errors */
  } result;

#include "types.h"
}

/* scanner and error count are exchanged between main, yyparse and yylex */
%param {yyscan_t scanner}{result *res}

%code provides /* -> header file, following the definition of YYSTYPE */
{
  /* the expected prototype of yylex:
   *   1st argument must be named `yylval_param'
   *   2nd argument must be named `yyscanner'
   *   3rd argument must be named `res'
   */
#define YY_DECL \
  enum yytokentype yylex ( \
	YYSTYPE* yylval_param, \
	yyscan_t yyscanner, result *res)
  YY_DECL;

  void yyerror (yyscan_t, result *, const char *, ...);
}

%code
{
static int power (int, int);
}

%define api.pure full
%define api.token.prefix {TOK_}
%define api.value.type union
%define parse.error verbose
%define parse.trace
%verbose

%token
  EQ     "="
  PLUS   "+"
  MINUS  "-"
  STAR   "*"
  SLASH  "/"
  UP     "^"
  LPAREN "("
  RPAREN ")"
  EOL    "end-of-line"
  EOF 0  "end-of-file"
;

%token <int> NUM "number"
%type <int> exp
%printer { fprintf (yyo, "%d", $$); } <int>

/* precedence (lowest to highest) and associativity */
%nonassoc "="		/* comparison */
%left "+" "-"
%left "*" "/"
%precedence UNARY	/* unary plus or minus */
%right "^"		/* exponentiation */

@attributes {semantic_value iVal;} exp "number"
@attributes {@void print;} line

%%
input	:	line
	|	input line
	;

line	:	exp EOL			@{
					  @decorate
					  @e line.print: exp.iVal;
					    {res->value = @exp.iVal@;
					     if (res->verbose)
						fprintf (stderr, "%d\n", @exp.iVal@);
					    }
					@}
	|	error EOL		{ yyyabort(); yyerrok; }
					@{ @m line.print; {} @}
//	|	EOL			@{ @m @line.print@ {} @} // TES: poor syntax errors (3) for this missing ';'
	|	EOL			@{ @m @line.print@; {} @}
	;

exp	:	"number"		@{ @i @exp.iVal@ = @"number".iVal@; @}
	|	exp "=" exp		@{ @i @exp.0.iVal@ = (@exp.1.iVal@ == @exp.2.iVal@); @}
	|	exp "+" exp		@{ @i @exp.0.iVal@ = (@exp.1.iVal@ + @exp.2.iVal@); @}
	|	exp "-" exp		@{ @i @exp.0.iVal@ = (@exp.1.iVal@ - @exp.2.iVal@); @}
	|	exp "*" exp		@{ @i @exp.0.iVal@ = (@exp.1.iVal@ * @exp.2.iVal@); @}
	|	exp "/" exp		{
					 if ($3 == 0)
					   {yyerror (scanner, res, "invalid division by zero");
					    YYERROR;
					   }
					}
					@{ @i @exp.0.iVal@ = (@exp.1.iVal@ / @exp.2.iVal@); @}
	|	"+" exp %prec UNARY	@{ @i @exp.0.iVal@ = + @exp.1.iVal@; @}
	|	"-" exp %prec UNARY	@{ @i @exp.0.iVal@ = - @exp.1.iVal@; @}
	|	exp "^" exp		@{ @i @exp.0.iVal@ = power (@exp.1.iVal@, @exp.2.iVal@); @}
	|	"(" exp ")"		@{ @i @exp.0.iVal@ = @exp.1.iVal@; @}
	;

%%

#include <assert.h>

static int
power (int base, int exponent)
{
  int res = 1;
  assert (0 <= exponent);
  for (/* Niente */; exponent; --exponent)
    res *= base;
  return res;
}

#include "lexer.h"

result
parse (FILE *infile)
{
  yyscan_t scanner;
  yylex_init (&scanner);
  yyset_in (infile, scanner);
  result res = {1, 0, 0};
  yyparse (scanner, &res);
  yylex_destroy (scanner);
  return res;
}

#include <stdio.h>
#include <stdarg.h>

void
yyerror (yyscan_t scanner, result *res,
         const char *msg, ...)
{
  (void) scanner; /* avoid unreferenced parameter warning */
  va_list args;
  va_start (args, msg);
  vfprintf (stderr, msg, args);
  va_end (args);
  fputc ('\n', stderr);
  ++res->nerrs;
}

#include <stdlib.h>

int
main (int argc, char *argv[])
{
 char *inputFilename = (char *) NULL;
 FILE *infile = stdin;

#if YYDEBUG
 /* enable parser runtime debugging */
 if (getenv("YYDEBUG") != 0) yydebug = 1;
#endif

 if (--argc >= 1) /* ignore the program name */
    inputFilename = argv[argc-1];

 if (inputFilename != (char *) NULL)
   {
    infile = fopen (inputFilename, "r");
    if (!infile)
      {perror (inputFilename);
       exit(-1);
      }
   }

 result res = parse (infile);
 if (res.nerrs) return 1; /* failure if there were errors */

 return 0;
}
